"""
Collects `Line of Route (LOR/PRIDE) <http://www.railwaycodes.org.uk/pride/pride0.shtm>`_ codes.
"""

import itertools
import os
import re
import urllib.parse

import bs4
import pandas as pd
from pyhelpers.ops import remove_dict_keys, update_dict_keys
from pyhelpers.text import find_similar_str

from .._base import _Base
from ..parser import _get_last_updated_date, parse_tr
from ..utils import fetch_all_verbose, home_page_url, print_inst_conn_err, print_void_msg


class LOR(_Base):
    """
    A class for collecting data of
    `Line of Route (LOR/PRIDE) <http://www.railwaycodes.org.uk/pride/pride0.shtm>`_.

    .. note::

        'LOR' and 'PRIDE' stands for 'Line Of Route' and 'Possession Resource Information Database',
        respectively.
    """

    #: The name of the data.
    NAME: str = 'Possession Resource Information Database (PRIDE)/Line Of Route (LOR) codes'
    #: A short name of the data
    SHORT_NAME: str = 'Line of Route (LOR/PRIDE) codes'
    #: The key for accessing the data.
    KEY: str = 'LOR'
    #: The key for accessing the data of *prefixes*.
    KEY_P: str = 'Key to prefixes'
    #: The key for accessing the data of *ELR/LOR converter*.
    KEY_ELC: str = 'ELR/LOR converter'

    #: The URL of the main web page for the data.
    URL = urllib.parse.urljoin(home_page_url(), '/pride/pride0.shtm')

    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the data catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.
        :ivar list valid_prefixes: A list of valid prefixes.

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> lor.NAME
            'Possession Resource Information Database (PRIDE)/Line Of Route (LOR) codes'
            >>> lor.URL
            'http://www.railwaycodes.org.uk/pride/pride0.shtm'
        """

        super().__init__(
            data_dir=data_dir, content_type='catalogue', data_category="line-data",
            update=update, verbose=verbose)

        self.valid_prefixes = self.get_keys_to_prefixes(prefixes_only=True)

    def validate_prefix(self, prefix):
        """
        Validates and standardises a PRIDE/LOR code prefix.

        If the provided `prefix` is not found in the list of valid prefixes, an attempt is made
        to find the closest matching valid prefix. If no match is found, an error is raised.

        :param prefix: The PRIDE/LOR code prefix to validate.
        :type prefix: str
        :raises AssertionError: If the `prefix` is not a valid PRIDE/LOR prefix.
        :return: A validated and standardised uppercase prefix.
        :rtype: str

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> lor.validate_prefix(prefix='cy')
            'CY'
            >>> lor.validate_prefix(prefix='ca')
            Traceback (most recent call last):
                ...
            AssertionError: `prefix` must be one of ['CY', 'EA', 'GW', 'LN', 'MD', 'NW', 'NZ', ...
        """

        prefix_ = prefix.upper()

        if prefix_ not in self.valid_prefixes:
            prefix_ = find_similar_str(prefix_, self.valid_prefixes)

            if not prefix_:
                raise AssertionError(f"`prefix` must be one of {self.valid_prefixes}")

        return prefix_

    def get_url(self, prefix):
        """
        Generates the URL for the given PRIDE/LOR code prefix.

        This method constructs the appropriate webpage URL based on the provided `prefix`,
        ensuring that it is valid before appending the correct suffix.

        :param prefix: The PRIDE/LOR code prefix.
        :type prefix: str
        :return: A fully constructed URL corresponding to the given prefix.
        :rtype: str

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> lor.get_url(prefix='CY')
            'http://www.railwaycodes.org.uk/pride/pridecy.shtm'
            >>> lor.get_url(prefix='CA')
            Traceback (most recent call last):
                ...
            AssertionError: `prefix` must be one of ['CY', 'EA', 'GW', 'LN', 'MD', 'NW', 'NZ', ...
        """

        url = urllib.parse.urljoin(home_page_url(), '/pride/pride')

        prefix_ = self.validate_prefix(prefix)

        if prefix_ in ("NW", "NZ"):
            url += 'nw.shtm'
        else:
            url += f'{prefix_.lower()}.shtm'

        return url

    def _parse_keys_to_prefixes(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')
        span_tags = soup.find_all(name='span', attrs={'class': 'tab2'})

        data = [
            (x.get_text(strip=True), str(x.next_sibling).strip().replace('=  ', ''))
            for x in span_tags]

        lor_pref = pd.DataFrame(data=data, columns=['Prefixes', 'Name'])

        keys_to_prefixes = {
            self.KEY_P: lor_pref, self.KEY_TO_LAST_UPDATED_DATE: self.last_updated_date}

        self._save_data_to_file(
            data=keys_to_prefixes, data_name="keys-to-prefixes", verbose=verbose)

        return keys_to_prefixes

    def collect_keys_to_prefixes(self, confirmation_required=True, verbose=False,
                                 raise_error=False):
        # noinspection PyShadowingNames
        """
        Collects the keys to PRIDE/LOR code prefixes from the source web page.

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing code prefixes and the last updated date,
            or ``None`` if no data is available.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> lor_page_urls = lor.collect_keys_to_prefixes()
            To collect data of URLs to LOR codes web pages
            ? [No]|Yes: yes
            Collecting the data ... Done.
            >>> lor_page_urls[0]
            'http://www.railwaycodes.org.uk/pride/pridecy.shtm'
        """

        keys_to_prefixes = self._collect_data_from_source(
            data_name='keys to LOR prefixes', method=self._parse_keys_to_prefixes, url=self.URL,
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return keys_to_prefixes

    def get_keys_to_prefixes(self, prefixes_only=True, update=False, dump_dir=None, verbose=False,
                             **kwargs):
        """
        Gets the keys to PRIDE/LOR code prefixes.

        :param prefixes_only: If ``True`` (default), returns only the prefixes;
            otherwise, additional information, including the last updated date.
        :type prefixes_only: bool
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A list of the keys to LOR code prefixes if ``prefixes_only=True``,
            otherwise a dictionary containing code prefixes and the last updated date,
            or ``None`` if no data is available.
        :rtype: list | dict | None

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> keys_to_pfx = lor.get_keys_to_prefixes()
            >>> keys_to_pfx
            ['CY', 'EA', 'GW', 'LN', 'MD', 'NW', 'NZ', 'SC', 'SO', 'SW', 'XR']
            >>> keys_to_pfx = lor.get_keys_to_prefixes(prefixes_only=False)
            >>> type(keys_to_pfx)
            dict
            >>> list(keys_to_pfx.keys())
            ['Key to prefixes', 'Last updated date']
            >>> keys_to_pfx_codes = keys_to_pfx['Key to prefixes']
            >>> type(keys_to_pfx_codes)
            pandas.core.frame.DataFrame
            >>> keys_to_pfx_codes.head()
              Prefixes                                    Name
            0       CY                                   Wales
            1       EA         South Eastern: East Anglia area
            2       GW  Great Western (later known as Western)
            3       LN                  London & North Eastern
            4       MD       North West: former Midlands lines
        """

        kwargs.update({'data_name': "keys-to-prefixes", 'method': self.collect_keys_to_prefixes})

        keys_to_prefixes = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        if prefixes_only:
            keys_to_prefixes = keys_to_prefixes[self.KEY_P]['Prefixes'].to_list()

            if update and len(keys_to_prefixes) > 0:
                self.valid_prefixes = keys_to_prefixes

        return keys_to_prefixes

    def _parse_page_urls(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        links = soup.find_all(
            name='a', href=re.compile('^pride|elrmapping'),
            string=re.compile('.*(codes|converter|Historical)'))

        lor_page_urls = list(
            dict.fromkeys([self.URL.replace(os.path.basename(self.URL), x['href']) for x in links]))

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(data=lor_page_urls, data_name="prefix-page-urls", verbose=verbose)

        return lor_page_urls

    def collect_page_urls(self, confirmation_required=True, verbose=False, raise_error=False):
        # noinspection PyShadowingNames
        """
        Collects a list of URLs to `PRIDE/LOR codes`_ web pages.

        .. _`PRIDE/LOR codes`: http://www.railwaycodes.org.uk/pride/pride0.shtm

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A list of URLs of web pages the LOR codes.
        :rtype: list

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> lor_page_urls = lor.collect_page_urls()
            To collect data of URLs to LOR codes web pages
            ? [No]|Yes: yes
            Collecting the data ... Done.
            >>> lor_page_urls[0]
            'http://www.railwaycodes.org.uk/pride/pridecy.shtm'
        """

        lor_page_urls = self._collect_data_from_source(
            data_name='URLs to LOR codes web pages', method=self._parse_page_urls, url=self.URL,
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return lor_page_urls

    def get_page_urls(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Gets a list of URLs to
        `PRIDE/LOR codes <http://www.railwaycodes.org.uk/pride/pride0.shtm>`_ web pages.

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A list of URLs of the web pages hosting LOR codes for each prefix.
        :rtype: list | None

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> lor_urls = lor.get_page_urls()
            >>> type(lor_urls)
            list
            >>> lor_urls[0]
            'http://www.railwaycodes.org.uk/pride/pridecy.shtm'
        """

        kwargs.update({'data_name': "prefix-page-urls", 'method': self.collect_page_urls})

        lor_page_urls = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return lor_page_urls

    @staticmethod
    def _parse_line_name(x):
        """
        Parses the column of 'Line Name'.

        :return: The line name and its corresponding note.
        :rtype: tuple
        """
        # re.search('\w+.*(?= \(\[\')', x).group()
        # re.search('(?<=\(\[\')\w+.*(?=\')', x).group()
        try:
            line_name, line_name_note = x.split(' ([\'')
            line_name_note = line_name_note.strip('\'])')

        except ValueError:
            line_name, line_name_note = x, ''

        return line_name, line_name_note

    @staticmethod
    def _parse_ra(x):
        # x = '3✖Originally reported as RA4'
        if '✖' in x:
            ra_dat, ra_note = x.split('✖')
        elif x == '[unknown]':
            ra_dat, ra_note = '', 'unknown'
        else:
            ra_dat, ra_note = x, ''

        return ra_dat, ra_note

    def _parse_h3_table(self, tbl):
        # thead, tbody = tbl[0].find('thead'), tbl[0].find('tbody')
        thead, tbody = tbl.find('thead'), tbl.find('tbody')
        ths = [x.text.replace('\n', ' ') for x in thead.find_all('th')]
        trs = tbody.find_all('tr')
        tbl_ = parse_tr(trs=trs, ths=ths, as_dataframe=True)

        ln_col_name, ra_col_name = 'Line Name', 'Route Availability (RA)'

        tbl_[[ln_col_name, ln_col_name + ' Note']] = pd.DataFrame(
            tbl_[ln_col_name].map(self._parse_line_name).to_list(), index=tbl_.index)

        tbl_[[ra_col_name, 'RA Note']] = pd.DataFrame(
            tbl_[ra_col_name].map(self._parse_ra).to_list(), index=tbl_.index)

        try:
            # note_dat_ = [
            #     (x['id'].title(), x.text.strip().replace('\xa0', ''))
            #     for x in soup.find('ol').findChildren('a')]
            # note_dat = dict(note_dat_)
            note = tbl.find_previous('p').text

        except AttributeError:
            # note = dict([('Note', None)])
            note = ''

        return tbl_, note

    def _parse_codes(self, prefix, source, verbose=False):
        # prefix = prefix.upper()
        # assert prefix in self.valid_prefixes, f"`prefix` must be one of {self.valid_prefixes}"

        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        h3, table = soup.find_all(name='h3'), soup.find_all(name='table')
        if len(h3) == 0:
            code_data, code_notes = self._parse_h3_table(table[0])

            lor_codes_by_initials = {prefix: code_data, 'Notes': code_notes}

        else:
            # code_data_and_notes = [
            #     dict(zip([prefix, 'Notes'], self._parse_h3_table(x, soup=soup)))
            #     for x in zip(*[iter(table)] * 2)]
            data_and_note = [self._parse_h3_table(tbl) for tbl in table]
            keys = [(h3_.text, h3_.text + ' note') for h3_ in h3]

            code_data_and_notes = dict(
                zip(*map(itertools.chain.from_iterable, [keys, data_and_note])))

            lor_codes_by_initials = {prefix: code_data_and_notes}

        last_updated_date = _get_last_updated_date(soup=soup)

        lor_codes_by_initials.update({self.KEY_TO_LAST_UPDATED_DATE: last_updated_date})

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=lor_codes_by_initials, data_name=prefix.lower(), sub_dir="prefixes",
            verbose=verbose)

        return lor_codes_by_initials

    def collect_codes(self, prefix, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects data of `PRIDE/LOR codes <http://www.railwaycodes.org.uk/pride/pride0.shtm>`_
        for the given prefix.

        :param prefix: The prefix of LOR codes to collect.
        :type prefix: str
        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the LOR codes for the given ``prefix``,
            or ``None`` if no data is available.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> lor_codes_cy = lor.collect_codes(prefix='CY')
            >>> type(lor_codes_cy)
            dict
            >>> list(lor_codes_cy.keys())
            ['CY', 'Notes', 'Last updated date']
            >>> cy_codes = lor_codes_cy['CY']
            >>> type(cy_codes)
            pandas.core.frame.DataFrame
            >>> cy_codes.head()
                 Code  ...                       RA Note
            0   CY240  ...           Caerwent branch RA4
            1  CY1540  ...  Pembroke - Pembroke Dock RA6
            [2 rows x 5 columns]
            >>> lor_codes_nw = lor.collect_codes(prefix='NW')
            >>> type(lor_codes_nw)
            dict
            >>> list(lor_codes_nw.keys())
            ['NW/NZ', 'Notes', 'Last updated date']
            >>> nw_codes = lor_codes_nw['NW/NZ']
            >>> nw_codes.head()
                 Code  ... RA Note
            0  NW1001  ...
            1  NW1002  ...
            2  NW1003  ...
            3  NW1004  ...
            4  NW1005  ...
            [5 rows x 5 columns]
            >>> lor_codes_xr = lor.collect_codes(prefix='XR')
            >>> type(lor_codes_xr)
            dict
            >>> list(lor_codes_xr.keys())
            ['XR', 'Last updated date']
            >>> xr_codes = lor_codes_xr['XR']
            >>> type(xr_codes)
            dict
            >>> list(xr_codes.keys())
            ['Current codes', 'Current codes note', 'Past codes', 'Past codes note']
            >>> xr_codes['Past codes'].head()
                Code  ... RA Note
            0  XR001  ...
            1  XR002  ...
            [2 rows x 5 columns]
            >>> xr_codes['Current codes'].head()
                Code  ...                     RA Note
            0  XR001  ...  Originally reported as RA4
            1  XR002  ...  Originally reported as RA4
            [2 rows x 5 columns]
        """

        prefix_ = self.validate_prefix(prefix=prefix)

        url = self.get_url(prefix=prefix_)

        lor_codes = self._collect_data_from_source(
            data_name=self.KEY, method=self._parse_codes, url=url, initial=prefix_, prefix=prefix_,
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return lor_codes

    def fetch_codes(self, prefix=None, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `PRIDE/LOR codes`_.

        .. _`PRIDE/LOR codes`: http://www.railwaycodes.org.uk/pride/pride0.shtm

        :param prefix: The prefix of LOR codes; defaults to ``None``.
        :type prefix: str | None
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the LOR codes.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> lor_codes_dat_cy = lor.fetch_codes(prefix='CY')
            >>> type(lor_codes_dat_cy)
            dict
            >>> list(lor_codes_dat_cy)
            ['CY', 'Notes', 'Last updated date']
            >>> lor_codes_dat_cy['CY']
                 Code  ...                       RA Note
            0   CY240  ...           Caerwent branch RA4
            1  CY1540  ...  Pembroke - Pembroke Dock RA6
            [2 rows x 5 columns]
            >>> lor_codes_dat = lor.fetch_codes()
            >>> type(lor_codes_dat)
            dict
            >>> list(lor_codes_dat.keys())
            ['LOR', 'Last updated date']
            >>> l_codes = lor_codes_dat['LOR']
            >>> type(l_codes)
            dict
            >>> list(l_codes.keys())[:5]
            ['CY', 'CY Notes', 'EA', 'EA Notes', 'GW']
            >>> cy_codes = l_codes['CY']
            >>> type(cy_codes)
            pandas.core.frame.DataFrame
            >>> cy_codes
                 Code  ...                       RA Note
            0   CY240  ...           Caerwent branch RA4
            1  CY1540  ...  Pembroke - Pembroke Dock RA6
            [2 rows x 5 columns]
            >>> xr_codes = l_codes['XR']
            >>> type(xr_codes)
            dict
            >>> list(xr_codes.keys())
            ['Current codes', 'Current codes note', 'Past codes', 'Past codes note']
            >>> xr_codes['Past codes']
                Code  ... RA Note
            0  XR001  ...
            1  XR002  ...
            [2 rows x 5 columns]
            >>> xr_codes['Current codes']
                Code  ...                     RA Note
            0  XR001  ...  Originally reported as RA4
            1  XR002  ...  Originally reported as RA4
            [2 rows x 5 columns]
        """

        if prefix:
            prefix_ = self.validate_prefix(prefix)
            args = {
                'data_name': prefix_.lower(),
                'method': self.collect_codes,
                'sub_dir': "prefixes",
                'prefix': prefix_,
            }
            kwargs.update(args)

            lor_data = self._fetch_data_from_file(
                update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        else:
            prefixes = self.get_keys_to_prefixes(prefixes_only=True, verbose=verbose)

            # # Adjust prefixes list: replace 'NW' with 'NW/NZ' and remove 'NZ'
            # prefixes_ = ['NW/NZ' if p == 'NW' else p for p in prefixes if p != 'NZ']

            # Set verbosity based on conditions
            verbose_ = fetch_all_verbose(data_dir=dump_dir, verbose=verbose)

            # Fetch LOR codes
            lor_codes = [
                self.fetch_codes(prefix=p, update=update, verbose=verbose_, **kwargs)
                for p in prefixes]

            # Retry if all fetches failed
            if all(x is None for x in lor_codes):
                if update:
                    print_inst_conn_err(verbose=verbose)
                    print_void_msg(data_name=self.KEY.lower(), verbose=verbose)

                lor_codes = [
                    self.fetch_codes(prefix=p, update=False, verbose=verbose_)
                    for p in prefixes]

            # Process fetched data
            lor_data, last_updated_dates = {self.KEY: {}}, []

            for p, dat in zip(prefixes, lor_codes):
                last_updated_dates.append(dat.get(self.KEY_TO_LAST_UPDATED_DATE))
                remove_dict_keys(dat, self.KEY_TO_LAST_UPDATED_DATE)

                # Merge fetched data with optional renaming of "Notes" keys
                lor_data[self.KEY].update(
                    update_dict_keys(dat, replacements={'Notes': f'{p} Notes'}) if 'Notes' in dat
                    else dat)

            # Get the latest updated date
            latest_update_date = max(filter(None, last_updated_dates), default=None)
            lor_data.update({self.KEY_TO_LAST_UPDATED_DATE: latest_update_date})

        if dump_dir:
            self._save_data_to_file(
                data=lor_data, data_name=self.KEY, dump_dir=dump_dir, verbose=verbose)

        return lor_data

    # == ELR/LOR converter =========================================================================

    def _collect_elr_lor_converter(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        thead, tbody = soup.find_all('thead')[0], soup.find_all('tbody')[0]
        ths = thead.find_all('th')
        trs = tbody.find_all(name='tr')

        elr_lor_dat = parse_tr(trs=trs, ths=ths, as_dataframe=True)

        elr_links = soup.find_all(name='td', string=re.compile(r'([A-Z]{3})(\d)?'))
        lor_links = soup.find_all(name='a', href=re.compile(r'pride([a-z]{2})\.shtm#'))

        # if len(elr_links) != len(elr_lor_dat):
        #     duplicates = elr_lor_dat[elr_lor_dat.duplicated(['ELR', 'LOR code'], keep=False)]
        #     for i in duplicates.index:
        #         if not duplicates['ELR'].loc[i].lower() in elr_links[i]:
        #             elr_links.insert(i, elr_links[i - 1])
        #         if not lor_links[i].endswith(duplicates['LOR code'].loc[i].lower()):
        #             lor_links.insert(i, lor_links[i - 1])

        elr_lor_dat['ELR_URL'] = [
            urllib.parse.urljoin(home_page_url(), x.a.get('href')) if x.a else None
            for x in elr_links]

        elr_lor_dat['LOR_URL'] = [
            urllib.parse.urljoin(home_page_url(), 'pride/' + x.get('href'))
            for x in lor_links]

        elr_lor_converter = {
            self.KEY_ELC: elr_lor_dat,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup)}

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=elr_lor_converter, data_name=re.sub(r"[/ ]", "-", self.KEY_ELC), verbose=verbose)

        return elr_lor_converter

    def collect_elr_lor_converter(self, confirmation_required=True, verbose=False,
                                  raise_error=False):
        """
        Collects data of
        `ELR/LOR converter <http://www.railwaycodes.org.uk/pride/elrmapping.shtm>`_
        from the source web page.

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of ELR/LOR converter,
            or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> elr_lor_conv = lor.collect_elr_lor_converter()
            To collect data of ELR/LOR converter
            ? [No]|Yes: yes
            >>> type(elr_lor_conv)
            dict
            >>> list(elr_lor_conv.keys())
            ['ELR/LOR converter', 'Last updated date']
            >>> elr_loc_conv_data = elr_lor_conv['ELR/LOR converter']
            >>> type(elr_loc_conv_data)
            pandas.core.frame.DataFrame
            >>> elr_loc_conv_data.head()
                ELR  ...                                            LOR_URL
            0   AAV  ...  http://www.railwaycodes.org.uk/pride/pridesw.s...
            1   ABD  ...  http://www.railwaycodes.org.uk/pride/pridegw.s...
            2   ABE  ...  http://www.railwaycodes.org.uk/pride/prideln.s...
            3  ABE1  ...  http://www.railwaycodes.org.uk/pride/prideln.s...
            4  ABE2  ...  http://www.railwaycodes.org.uk/pride/prideln.s...
            [5 rows x 6 columns]
        """

        elr_lor_converter = self._collect_data_from_source(
            data_name=self.KEY_ELC, method=self._collect_elr_lor_converter,
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return elr_lor_converter

    def fetch_elr_lor_converter(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `ELR/LOR converter`_.

        .. _`ELR/LOR converter`: http://www.railwaycodes.org.uk/pride/elrmapping.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of ELR/LOR converter.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LOR  # from pyrcs import LOR
            >>> lor = LOR()
            >>> elr_lor_conv = lor.fetch_elr_lor_converter()
            >>> type(elr_lor_conv)
            dict
            >>> list(elr_lor_conv.keys())
            ['ELR/LOR converter', 'Last updated date']
            >>> elr_loc_conv_data = elr_lor_conv['ELR/LOR converter']
            >>> type(elr_loc_conv_data)
            pandas.core.frame.DataFrame
            >>> elr_loc_conv_data.head()
                ELR  ...                                            LOR_URL
            0   AAV  ...  http://www.railwaycodes.org.uk/pride/pridesw.s...
            1   ABD  ...  http://www.railwaycodes.org.uk/pride/pridegw.s...
            2   ABE  ...  http://www.railwaycodes.org.uk/pride/prideln.s...
            3  ABE1  ...  http://www.railwaycodes.org.uk/pride/prideln.s...
            4  ABE2  ...  http://www.railwaycodes.org.uk/pride/prideln.s...
            [5 rows x 6 columns]
        """

        args = {
            'data_name': re.sub(r"[/ ]", "-", self.KEY_ELC),
            'method': self.collect_elr_lor_converter,
        }
        kwargs.update(args)

        elr_lor_converter = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return elr_lor_converter



"""
Package initialisation.
"""

import datetime
import json
import pkgutil

from .collector import LineData, OtherAssets
from .line_data import Bridges, ELRMileages, Electrification, LOR, LineNames, LocationIdentifiers, \
    TrackDiagrams
from .other_assets import Buzzer, Depots, Features, HabdWild, SignalBoxes, Stations, Telegraph, \
    Tunnels, Viaducts, WaterTroughs

metadata = json.loads(pkgutil.get_data(__name__, "data/.metadata").decode())

__project__ = metadata['Project']
__pkgname__ = metadata['Package']
__desc__ = metadata['Description']

__author__ = metadata['Author']
__affil__ = metadata['Affiliation']
__email__ = metadata['Email']

__version__ = metadata['Version']
__license__ = metadata['License']
__copyright__ = f'2019-{datetime.datetime.now().year}, {__author__}'

__first_release__ = metadata['First release']

__all__ = [
    'LineData',
    'OtherAssets',
    'Electrification',
    'ELRMileages',
    'LineNames',
    'LocationIdentifiers',
    'LOR',
    'TrackDiagrams',
    'Bridges',
    'SignalBoxes',
    'Tunnels',
    'Viaducts',
    'Stations',
    'Depots',
    'Features',
    'HabdWild',
    'WaterTroughs',
    'Telegraph',
    'Buzzer',
]



"""
Collects data of `railway tunnel lengths <http://www.railwaycodes.org.uk/tunnels/tunnels0.shtm>`_.
"""

import itertools
import re
import urllib.parse

import bs4
import numpy as np
import pandas as pd

from .._base import _Base
from ..parser import _get_last_updated_date, parse_tr
from ..utils import home_page_url, is_home_connectable, print_inst_conn_err, print_void_msg, \
    validate_page_name


class Tunnels(_Base):
    """
    A class for collecting data of
    `railway tunnel lengths <http://www.railwaycodes.org.uk/tunnels/tunnels0.shtm>`_.
    """

    #: The name of the data.
    NAME: str = 'Railway tunnel lengths'
    #: The key for accessing the data.
    KEY: str = 'Tunnels'

    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(home_page_url(), '/tunnels/tunnels0.shtm')

    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.other_assets import Tunnels  # from pyrcs import Tunnels
            >>> tunl = Tunnels()
            >>> tunl.NAME
            'Railway tunnel lengths'
            >>> tunl.URL
            'http://www.railwaycodes.org.uk/tunnels/tunnels0.shtm'
        """

        super().__init__(
            data_dir=data_dir, content_type='catalogue', data_category="other-assets",
            update=update, verbose=verbose)

        self.page_range = range(1, 5)

    @staticmethod
    def _parse_length(x):
        """
        Parses data in ``'Length'`` column, i.e. convert miles/yards to metres.

        :param x: Raw length data.
        :type x: str | None
        :return: The parsed length data and, if any, additional information associated with it.
        :rtype: tuple

        **Examples**::

            >>> from pyrcs.other_assets import Tunnels  # from pyrcs import Tunnels
            >>> tunl = Tunnels()
            >>> tunl._parse_length('')
            (nan, 'Unavailable')
            >>> tunl._parse_length('1m 182y')
            (1775.7648, None)
            >>> tunl._parse_length('formerly 0m236y')
            (215.7984, 'Formerly')
            >>> tunl._parse_length('0.325km (0m 356y)')
            (325.5264, '0.325km')
            >>> tunl._parse_length("0m 48yd- (['0m 58yd'])")
            (48.4632, '43.89-53.04 metres')
        """

        if '✖' in x:
            x, note_ = x.split('✖')
            note_ = ' (' + note_ + ')'
        else:
            note_ = ''

        if re.match(r'[Uu]nknown', x):
            length = np.nan
            note = 'Unknown'

        elif x == '':
            length = np.nan
            note = 'Unavailable'

        elif re.match(r'\d+m \d+yd?(- | to )?.*\d+m \d+yd?.*', x):
            miles_a, yards_a, miles_b, yards_b = re.findall(r'\d+', x)
            length_a = float(miles_a) * 1609.344 + float(yards_a) * 0.9144
            # measurement.measures.Distance(mi=miles_a).m + measurement.measures.Distance(yd=yards_a).m
            length_b = float(miles_b) * 1609.344 + float(yards_b) * 0.9144
            # measurement.measures.Distance(mi=miles_b).m + measurement.measures.Distance(yd=yards_b).m
            length = (length_a + length_b) / 2
            note = '-'.join([str(round(length_a, 2)), str(round(length_b, 2))]) + ' metres'

        else:
            if re.match(r'(formerly )?c?≈?\d+m ?\d+yd?|ch', x):
                miles, yards = re.findall(r'\d+', x)
                if re.match(r'.*\d+ch$', x):  # "yards" is "chains"
                    yards = yards * 22  # measurement.measures.Distance(chain=yards).yd

                if re.match(r'^c.*|^≈', x):
                    note = 'Approximate'
                elif re.match(r'\d+y$', x):
                    note = re.search(r'(?<=\dy).*$', x).group(0)
                elif re.match(r'^(formerly).*', x):
                    note = 'Formerly'
                else:
                    note = ''

            elif re.match(r'\d+\.\d+km(\r)? .*(\[\')?\(\d+m \d+yd?\).*', x):
                miles, yards = re.findall(r'\d+', re.search(r'(?<=\()\d+.*(?=\))', x).group(0))
                note = re.search(r'.+(?= (\[\')?\()', x.replace('\r', '')).group(0)

            else:
                miles, yards = 0, 0
                note = ''

            length = float(miles) * 1609.344 + float(yards) * 0.9144
            # measurement.measures.Distance(mi=miles).m + measurement.measures.Distance(yd=yards).m

            if note != '':
                note = note + note_

        return length, note

    def _collect_codes(self, page_no, source, verbose=False):
        page_name = validate_page_name(self, page_no, valid_page_no=self.page_range)

        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        theads, tbodies = soup.find_all('thead'), soup.find_all('tbody')

        codes_dat = []
        for thead, tbody in zip(theads, tbodies):
            ths = [th.text.strip() for th in thead.find_all('th')]
            trs = tbody.find_all('tr')
            dat = parse_tr(trs=trs, ths=ths, as_dataframe=True)

            temp = [re.match('^Between.*', x) or x == '' for x in dat.columns]
            if bool(temp):
                indices = list(itertools.compress(range(len(temp)), temp))
                if len(indices) == 2:
                    new_cols = ['Station A', 'Station B']
                    repl_col_names = dict(zip(dat.columns[indices].to_list(), new_cols))
                    dat.rename(columns=repl_col_names, inplace=True)

            codes_dat.append(dat)

        len_cols = ['Length (metres)', 'Length (note)']
        for i, dat in enumerate(codes_dat):
            temp = dat['Length'].map(self._parse_length)
            codes_dat[i][len_cols] = pd.DataFrame(zip(*temp)).T

        if len(codes_dat) == 1:
            codes_ = codes_dat[0]
        else:
            codes_ = dict(zip([x.text.strip() for x in soup.find_all('h3')], codes_dat))

        codes_on_page = {
            page_name: codes_,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        data_name = re.sub(r"[()]", "", re.sub(r"[ -]", "-", page_name)).lower()
        self._save_data_to_file(data=codes_on_page, data_name=data_name, verbose=verbose)

        return codes_on_page

    def collect_codes(self, page_no, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects data of `railway tunnel lengths`_ for a specific page number
        from the source web page.

        .. _`railway tunnel lengths`: http://www.railwaycodes.org.uk/tunnels/tunnels0.shtm

        :param page_no: The page number to collect data from;
            valid values are ``1``, ``2``, ``3`` and ``4``.
        :type page_no: int | str
        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of tunnel lengths for the specified ``page_no``
            and the date of when the data was last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import Tunnels  # from pyrcs import Tunnels
            >>> tunl = Tunnels()
            >>> page_1 = tunl.collect_codes(page_no=1)
            >>> type(page_1)
            dict
            >>> list(page_1.keys())
            ['Page 1 (A-F)', 'Last updated date']
            >>> page_1_codes = page_1['Page 1 (A-F)']
            >>> type(page_1_codes)
            pandas.core.frame.DataFrame
            >>> page_1_codes.head()
                         Name  Other names, remarks  ... Length (metres) Length (note)
            0    Abbotscliffe                        ...       1775.7648
            1      Abercanaid           see Merthyr  ...             NaN   Unavailable
            2     Aberchalder         see Loch Oich  ...             NaN   Unavailable
            3  Aberdovey No 1  also called Frongoch  ...        182.8800
            4  Aberdovey No 2    also called Morfor  ...        200.2536
            [5 rows x 10 columns]
            >>> page_4 = tunl.collect_codes(page_no=4)
            >>> type(page_4)
            dict
            >>> list(page_4.keys())
            ['Page 4 (others)', 'Last updated date']
            >>> page_4_codes = page_4['Page 4 (others)']
            >>> type(page_4_codes)
            dict
            >>> list(page_4_codes.keys())
            ['Tunnels on industrial and other minor lines',
             'Large bridges that are not officially tunnels but could appear to be so']
            >>> key1 = 'Tunnels on industrial and other minor lines'
            >>> page_4_dat = page_4_codes[key1]
            >>> type(page_4_dat)
            pandas.core.frame.DataFrame
            >>> page_4_dat.head()
                                  Name Other names, remarks  ... Length (metres) Length (note)
            0             Ashes Quarry                       ...         56.6928
            1        Ashey Down Quarry                       ...         33.8328
            2  Baileycroft Quarry No 1                       ...         28.3464
            3  Baileycroft Quarry No 2                       ...         21.0312
            4            Basfords Hill                       ...         46.6344
            [5 rows x 6 columns]
            >>> key2 = 'Large bridges that are not officially tunnels but could appear to be so'
            >>> page_4_dat_ = page_4_codes[key2]
            >>> type(page_4_dat_)
            pandas.core.frame.DataFrame
            >>> page_4_dat_.head()
                            Name Other names, remarks  ... Length (metres) Length (note)
            0  A470/A472 (north)                       ...         35.6616
            1  A470/A472 (south)                       ...         28.3464
            2               A720                       ...        145.3896
            3                 A9        Aberdeen line  ...        141.7320
            4                 A9           Perth line  ...        146.3040
            [5 rows x 8 columns]
        """

        data_name = self.NAME.lower()

        page_name = validate_page_name(self, page_no, valid_page_no=self.page_range)

        data = self._collect_data_from_source(
            data_name=data_name,
            method=self._collect_codes, initial=page_name, page_no=page_no,
            url=self.catalogue.get(page_name), confirmation_required=confirmation_required,
            confirmation_prompt=f"To collect data of {data_name} ({page_name})\n?",
            verbose=verbose, raise_error=raise_error)

        return data

    def fetch_codes(self, page_no=None, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `railway tunnel lengths`_.

        .. _`railway tunnel lengths`: http://www.railwaycodes.org.uk/tunnels/tunnels0.shtm

        :param page_no: The page number to collect data from;
            valid values are ``1``, ``2``, ``3`` and ``4``; defaults to ``None``.
        :type page_no: int | str
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of tunnel lengths
            (including the name, length, owner and relative location) and
            the date of when the data was last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import Tunnels  # from pyrcs import Tunnels
            >>> tunl = Tunnels()
            >>> tunl_len_codes = tunl.fetch_codes()
            >>> type(tunl_len_codes)
            dict
            >>> list(tunl_len_codes.keys())
            ['Tunnels', 'Last updated date']
            >>> tunl.KEY
            'Tunnels'
            >>> tunl_len_codes_dat = tunl_len_codes[tunl.KEY]
            >>> type(tunl_len_codes_dat)
            dict
            >>> list(tunl_len_codes_dat.keys())
            ['Page 1 (A-F)', 'Page 2 (G-P)', 'Page 3 (Q-Z)', 'Page 4 (others)']
            >>> page_1_codes = tunl_len_codes_dat['Page 1 (A-F)']
            >>> type(page_1_codes)
            pandas.core.frame.DataFrame
            >>> page_1_codes.head()
                         Name  Other names, remarks  ... Length (metres) Length (note)
            0    Abbotscliffe                        ...       1775.7648
            1      Abercanaid           see Merthyr  ...             NaN   Unavailable
            2     Aberchalder         see Loch Oich  ...             NaN   Unavailable
            3  Aberdovey No 1  also called Frongoch  ...        182.8800
            4  Aberdovey No 2    also called Morfor  ...        200.2536
            [5 rows x 10 columns]
            >>> page_4_codes = tunl_len_codes_dat['Page 4 (others)']
            >>> type(page_4_codes)
            dict
            >>> list(page_4_codes.keys())
            ['Tunnels on industrial and other minor lines',
             'Large bridges that are not officially tunnels but could appear to be so']
            >>> key1 = 'Tunnels on industrial and other minor lines'
            >>> page_4_dat = page_4_codes[key1]
            >>> type(page_4_dat)
            pandas.core.frame.DataFrame
            >>> page_4_dat.head()
                                  Name Other names, remarks  ... Length (metres) Length (note)
            0             Ashes Quarry                       ...         56.6928
            1        Ashey Down Quarry                       ...         33.8328
            2  Baileycroft Quarry No 1                       ...         28.3464
            3  Baileycroft Quarry No 2                       ...         21.0312
            4            Basfords Hill                       ...         46.6344
            [5 rows x 6 columns]
            >>> key2 = 'Large bridges that are not officially tunnels but could appear to be so'
            >>> page_4_dat_ = page_4_codes[key2]
            >>> type(page_4_dat_)
            pandas.core.frame.DataFrame
            >>> page_4_dat_.head()
                            Name Other names, remarks  ... Length (metres) Length (note)
            0  A470/A472 (north)                       ...         35.6616
            1  A470/A472 (south)                       ...         28.3464
            2               A720                       ...        145.3896
            3                 A9        Aberdeen line  ...        141.7320
            4                 A9           Perth line  ...        146.3040
            [5 rows x 8 columns]
        """

        if page_no:
            page_name = validate_page_name(self, page_no, valid_page_no=self.page_range)

            args = {
                'data_name': re.sub(r"[()]", "", re.sub(r"[ -]", "-", page_name)).lower(),
                'method': self.collect_codes,
                'page_no': page_no,
            }
            kwargs.update(args)

            tunnel_lengths = self._fetch_data_from_file(
                update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        else:
            verbose_1 = False if (dump_dir or not verbose) else (2 if verbose == 2 else True)
            verbose_2 = verbose_1 if is_home_connectable() else False

            codes_on_pages = [
                self.fetch_codes(x, update=update, verbose=verbose_2) for x in self.page_range]

            if all(x is None for x in codes_on_pages):
                if update:
                    print_inst_conn_err(verbose=verbose)
                    print_void_msg(data_name=self.KEY, verbose=verbose)

                codes_on_pages = [
                    self.fetch_codes(x, update=False, verbose=verbose_1) for x in self.page_range]

            tunnel_lengths = {
                self.KEY: {next(iter(x)): next(iter(x.values())) for x in codes_on_pages},
                self.KEY_TO_LAST_UPDATED_DATE:
                    max(next(itertools.islice(iter(x.values()), 1, 2)) for x in codes_on_pages),
            }

        if dump_dir is not None:
            self._save_data_to_file(
                data=tunnel_lengths, data_name=self.KEY, dump_dir=dump_dir, verbose=verbose)

        return tunnel_lengths



"""
Collects `railway station data <http://www.railwaycodes.org.uk/stations/station0.shtm>`_.
"""

import re
import string
import urllib.parse

import bs4
import pandas as pd
from pyhelpers.text import remove_punctuation

from .._base import _Base
from ..parser import _get_last_updated_date, get_catalogue, parse_tr
from ..utils import cd_data, collect_in_fetch_verbose, home_page_url, is_home_connectable, \
    print_inst_conn_err, print_void_msg, validate_initial


def _split_elr_mileage_column(dat):
    if 'ELRMileage' in dat.columns:
        temp = dat['ELRMileage'].str.split(r'\t\t / | / \[', n=1, regex=True, expand=True)
        temp.columns = ['ELR', 'Mileage']

        # elr_dat = temp['ELR'].str.strip().str.replace(' ', ' &&& ')
        # elr_dat = temp['ELR'].str.strip().str.replace('  /  ', ' &&& ')

        def _to_repl(x, repl=' &&& '):
            if '  /  ' in x:
                y = x.replace('  /  ', repl)
            elif ' / ' in x:
                y = x.replace(' / ', repl)
            elif ' ' in x:
                y = x.replace(' ', repl)
            else:
                y = x
            return y

        elr_dat = temp['ELR'].str.strip().map(_to_repl)

        temp['Mileage'] = temp['Mileage'].str.split('\t\t / ').fillna('').map(
            lambda x: [' / '.join(x_.strip(' []').split('  ')) for x_ in x])
        mil_dat = temp['Mileage'].map(lambda x: x[0] if len(x) == 1 else ' &&& '.join(x))

        dat.drop(columns=['ELRMileage'], inplace=True)
        dat.insert(1, 'ELR', elr_dat)
        dat.insert(2, 'Mileage', mil_dat)

    return dat


def _check_row_spans(dat):
    """
    Checks data in which there are row spans.

    :param dat: Preprocessed data of the station locations.
    :type dat: pandas.DataFrame
    :return: Data with row spans (if there is any).
    :rtype: pandas.DataFrame
    """

    # temp0 = dat['Degrees Longitude'].str.split(' / ')
    temp0 = dat['Degrees Longitude'].str.split(r' / |\r', regex=True)
    temp1 = dat[temp0.map(lambda x: True if len(x) > 1 else False)]

    cols = ['ELR', 'Mileage', 'Degrees Longitude', 'Degrees Latitude', 'Grid Reference']
    cols_ = [x for x in temp1.columns if x not in cols]

    temp_dat = []
    for col in cols:
        temp2 = temp1[col].str.split(r' &&& |\r| / ', regex=True)

        if cols.index(col) == 0:
            temp_dat_ = temp1[cols_].join(temp2).explode(col, ignore_index=True)
        else:
            temp_dat_ = temp2.explode(ignore_index=True).to_frame(name=col)

        temp_dat.append(temp_dat_)

    temp_data = pd.concat(temp_dat, axis=1)
    temp_data = temp_data[dat.columns.to_list()]

    dat0 = pd.concat([dat.drop(index=temp1.index), temp_data], axis=0, ignore_index=True)

    dat0[['ELR', 'Mileage']] = dat0[['ELR', 'Mileage']].map(lambda x: x.split(' &&& '))
    dat0 = dat0.explode(['ELR', 'Mileage'], ignore_index=True)

    dat0.sort_values(['Station'], ignore_index=True, inplace=True)

    return dat0


def _parse_coordinates_columns(dat):
    """
    Parses ``'Degrees Longitude'`` and ``'Degrees Latitude'`` of the station locations data.

    :param dat: Preprocessed data of the station locations.
    :type dat: pandas.DataFrame
    :return: Data with parsed coordinates.
    :rtype: pandas.DataFrame
    """

    ll_col_names = ['Degrees Longitude', 'Degrees Latitude']

    dat[ll_col_names] = dat[ll_col_names].map(
        lambda x: None if x.strip() == '' else float(re.sub(r'(c\.)|≈', '', x)))

    return dat


def _parse_station_column(dat):
    """
    Parses ``'Station'`` of the station locations data.

    :param dat: Preprocessed data of the station locations.
    :type dat: pandas.DataFrame
    :return: Data with parsed station names and their corresponding CRS.
    :rtype: pandas.DataFrame

    **Tests**::

        x = 'Hythe Road\t\t / [CRS awaited]'
        x = 'Heathrow Junction [sometimes referred to as Heathrow Interchange]\t\t / [no CRS?]'
    """

    temp1 = dat['Station'].str.split('\t\t', n=1, expand=True)
    temp1.columns = ['Station', 'CRS']
    dat['Station'] = temp1['Station'].str.rstrip(' / ').str.strip()

    # Get notes for stations
    stn_note_ = pd.Series('', index=dat.index)
    for i, x in enumerate(temp1['Station']):
        if '[' in x and ']' in x:
            y = re.search(r' \[(.*)](✖.*)?', x).group(0)  # Station Note
            dat.loc[i, 'Station'] = x.replace(y, '').strip()
            if '✖' in y:
                stn_note_[i] = '; '.join([y_.strip(' []') for y_ in y.split('✖')])
            else:
                stn_note_[i] = y.strip(' []')

    dat.insert(loc=dat.columns.get_loc('Station') + 1, column='Station Note', value=stn_note_)

    temp2 = temp1['CRS'].str.replace(' / /', ' &&& ').str.split(
        r'  | / ', regex=True, expand=True).fillna('')

    if temp2.shape[1] == 1:
        temp2.columns = ['CRS']
        temp2 = pd.concat(
            [temp2, pd.DataFrame('', index=temp2.index, columns=['CRS Note'])], axis=1)
    else:
        temp2.columns = ['CRS', 'CRS Note']
        temp2['CRS Note'] = temp2['CRS Note'].str.strip('[]')

    temp2['CRS'] = temp2['CRS'].str.replace(r'[()]', '', regex=True).map(
        lambda z: ' and '.join(['{} [{}]'.format(*z_.split('✖')) for z_ in z.split(' &&& ')])
        if ' &&& ' in z else z).str.strip()

    dat = pd.concat([dat, temp2], axis=1)

    return dat


def _parse_owner_and_operator(x):
    """
    x = dat['Owner'][0]
    x = dat['Owner'][1]
    """

    if ' / and / ' in x:
        y, y_ = x.replace(' / and / ', ' &&& '), ''

    elif ' / ' in x or '\r' in x:
        x_ = re.split(r' / |\r', x)

        # y - Owners or operators; y_ - Former owners or operators
        if len(x_) > 1:
            y = x_[0]
            y_ = x_[1] if len(x_[1:]) == 1 else ' / '.join(x_[1:])
        else:
            y, y_ = x_[0], ''

    else:
        y, y_ = x, ''

    if '✖' in y and ' &&& ' in y:
        y = ' and '.join(['{} [{}]'.format(*z.split('✖')) for z in y.split(' &&& ')])

    if ' [from' in y:
        y = remove_punctuation(y)

    return y, y_


def _parse_owner_and_operator_columns(dat):
    """
    Parses ``'Owner'`` and ``'Operator'`` of the station locations data.

    :param dat: Preprocessed data of the station locations.
    :type dat: pandas.DataFrame
    :return: Data with parsed information of owners and operators.
    :rtype: pandas.DataFrame
    """

    owner_operator = []
    for col in ['Owner', 'Operator']:
        temp = pd.DataFrame(
            dat[col].map(_parse_owner_and_operator).to_list(), columns=[col, 'Former ' + col],
            index=dat.index)

        del dat[col]

        owner_operator.append(temp)

    dat = pd.concat([dat] + owner_operator, axis=1)

    return dat


class Stations(_Base):
    """
    A class for collecting
    `railway station data <http://www.railwaycodes.org.uk/stations/station0.shtm>`_.
    """

    #: The name of the data.
    NAME: str = 'Railway station data'
    #: The key for accessing the data.
    KEY: str = 'Stations'
    #: The key for accessing the data of *Mileages, operators and grid coordinates*.
    KEY_TO_STN: str = 'Mileages, operators and grid coordinates'

    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(home_page_url(), '/stations/station0.shtm')

    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.other_assets import Stations  # from pyrcs import Stations
            >>> stn = Stations()
            >>> stn.NAME
            'Railway station data'
            >>> stn.URL
            'http://www.railwaycodes.org.uk/stations/station0.shtm'
        """

        super().__init__(
            data_dir=data_dir, data_category="other-assets", update=update, verbose=verbose)

        self.catalogue = self.fetch_catalogue(update=update, verbose=False)

        self.station_names_errata = {
                "-By-": "-by-",
                "-In-": "-in-",
                "-En-Le-": "-en-le-",
                "-La-": "-la-",
                "-Le-": "-le-",
                "-On-": "-on-",
                "-The-": "-the-",
                " Of ": " of ",
                "-Super-": "-super-",
                "-Upon-": "-upon-",
                "-Under-": "-under-",
                "-Y-": "-y-",
            }

    def _collect_catalogue(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.text, features='html.parser')

        cold_soup = soup.find_all('nav')[1]

        hot_soup = {
            a.text: urllib.parse.urljoin(self.URL, a.get('href'))
            for a in cold_soup.find_all('a')}

        catalogue = {}
        for k, v in hot_soup.items():
            sub_cat = get_catalogue(url=v, update=True, json_it=False)

            if sub_cat != hot_soup:
                if k in sub_cat.keys():
                    sub_cat.pop(k)
                elif 'Introduction' in sub_cat.keys():
                    sub_cat.pop('Introduction')
                if v in sub_cat.values():
                    catalogue[k] = sub_cat
                else:
                    catalogue[k] = {'Introduction': v, **sub_cat}

            else:
                catalogue.update({k: v})

        if verbose in {True, 1}:
            print("Done.")

        data_name = urllib.parse.urlparse(
            self.URL).path.lstrip('/').rstrip('.shtm').replace('/', '-')
        self._save_data_to_file(
            data=catalogue, data_name=data_name, ext=".json", dump_dir=cd_data("catalogue"),
            verbose=verbose, indent=4)

        return catalogue

    def collect_catalogue(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects the catalogue of `railway station data`_ from the source web page.

        .. _`railway station data`: http://www.railwaycodes.org.uk/stations/station0.shtm

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the catalogue of railway station data,
            or ``None`` if no data catalogue is collected.
        :rtype: dict | None
        """

        catalogue = self._collect_data_from_source(
            data_name=f'{self.NAME.lower()} catalogue', method=self._collect_catalogue,
            url=self.URL, confirmation_required=confirmation_required, verbose=verbose,
            raise_error=raise_error)

        return catalogue

    def fetch_catalogue(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches the catalogue of `railway station data`_.

        .. _`railway station data`: http://www.railwaycodes.org.uk/stations/station0.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the catalogue of railway station data,
            or ``None`` if no data catalogue is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.other_assets import Stations  # from pyrcs import Stations
            >>> stn = Stations()
            >>> stn_data_cat = stn.fetch_catalogue()
            >>> type(stn_data_cat)
            dict
            >>> list(stn_data_cat.keys())
            ['Mileages, operators and grid coordinates',
             'Bilingual names',
             'Sponsored signs',
             'Not served by SFO',
             'International',
             'Trivia',
             'Access rights',
             'Barrier error codes',
             'London Underground',
             'Railnet']
        """

        data_name = urllib.parse.urlparse(
            self.URL).path.lstrip('/').rstrip('.shtm').replace('/', '-')

        args = {
            'data_name': data_name,
            'method': self.collect_catalogue,
            'ext': ".json",
            'data_dir': cd_data("catalogue"),
        }
        kwargs.update(args)

        catalogue = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return catalogue

    def _collect_locations(self, initial, source, verbose=False):
        initial_ = validate_initial(initial)

        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')
        thead, tbody = soup.find('thead'), soup.find('tbody')

        # Create a DataFrame of the requested table
        trs = tbody.find_all(name='tr')
        ths = [re.sub(r'\n?\r+\n?', ' ', h.text).strip() for h in thead.find_all('th')]
        dat_ = parse_tr(trs=trs, ths=ths, as_dataframe=True)

        dat = dat_.copy()

        parser_funcs = [
            _split_elr_mileage_column,
            _check_row_spans,
            _parse_coordinates_columns,
            _parse_station_column,
            _parse_owner_and_operator_columns,
        ]
        for parser_func in parser_funcs:
            dat = parser_func(dat)

        # # Debugging
        # for parser_func in parser_funcs:
        #     try:
        #         dat_ = parser_func(dat_)
        #     except Exception:
        #         print(parser_func)
        #         break

        # Explode by ELR and Mileage
        dat = dat.explode(column=['ELR', 'Mileage'], ignore_index=True)

        # dat['Station'].replace(self.station_names_errata, regex=True, inplace=True)
        dat['Station'] = dat['Station'].replace(self.station_names_errata, regex=True)

        data = {
            initial_: dat.sort_values('Station', ignore_index=True),
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup)
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(data=data, data_name=initial_, sub_dir="a-z", verbose=verbose)

        return data

    def collect_locations(self, initial, confirmation_required=True, verbose=False,
                          raise_error=False):
        """
        Collects data of `railway station locations
        <http://www.railwaycodes.org.uk/stations/station0.shtm>`_
        (mileages, operators and grid coordinates) for a given initial letter.

        :param initial: The initial letter (e.g. ``'a'``, ``'z'``) of railway station names.
        :type initial: str
        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of railway station locations whose initial letters
            are the given ``initial`` and date of when the data was last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import Stations  # from pyrcs import Stations
            >>> stn = Stations()
            >>> stn_loc_a_codes = stn.collect_locations(initial='a')
            >>> type(stn_loc_a_codes)
            dict
            >>> list(stn_loc_a_codes.keys())
            ['A', 'Last updated date']
            >>> stn_loc_a_codes_dat = stn_loc_a_codes['A']
            >>> type(stn_loc_a_codes_dat)
            pandas.core.frame.DataFrame
            >>> stn_loc_a_codes_dat.head()
                  Station  ...                                    Former Operator
            0  Abbey Wood  ...  London & South Eastern Railway from 1 April 20...
            1  Abbey Wood  ...  London & South Eastern Railway from 1 April 20...
            2        Aber  ...  Keolis Amey Operations/Gweithrediadau Keolis A...
            3   Abercynon  ...  Keolis Amey Operations/Gweithrediadau Keolis A...
            4   Abercynon  ...  Keolis Amey Operations/Gweithrediadau Keolis A...
            [5 rows x 14 columns]
            >>> stn_loc_a_codes_dat.columns.to_list()
            ['Station',
             'Station Note',
             'ELR',
             'Mileage',
             'Status',
             'Degrees Longitude',
             'Degrees Latitude',
             'Grid Reference',
             'CRS',
             'CRS Note',
             'Owner',
             'Former Owner',
             'Operator',
             'Former Operator']
            >>> stn_loc_a_codes_dat[['Station', 'ELR', 'Mileage']].head()
                  Station  ELR   Mileage
            0  Abbey Wood  NKL  11m 43ch
            1  Abbey Wood  XRS  24.458km
            2        Aber  CAR   8m 69ch
            3   Abercynon  CAM  16m 28ch
            4   Abercynon  ABD  16m 28ch
        """

        initial_ = validate_initial(x=initial)

        data = self._collect_data_from_source(
            data_name=self.KEY_TO_STN.lower(), method=self._collect_locations, initial=initial_,
            url=self.catalogue[self.KEY_TO_STN].get(initial_),
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return data

    def fetch_locations(self, initial=None, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `railway station locations`_ (mileages, operators and grid coordinates).

        .. _`railway station locations`:
            http://www.railwaycodes.org.uk/stations/station0.shtm

        :param initial: The initial letter (e.g. ``'a'``, ``'z'``) of railway station names.
        :type initial: str
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of railway station locations and
            the date of when the data was last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import Stations  # from pyrcs import Stations
            >>> stn = Stations()
            >>> stn_loc_codes = stn.fetch_locations()
            >>> type(stn_loc_codes)
            dict
            >>> list(stn_loc_codes.keys())
            ['Mileages, operators and grid coordinates', 'Last updated date']
            >>> stn.KEY_TO_STN
            'Mileages, operators and grid coordinates'
            >>> stn_loc_codes_dat = stn_loc_codes[stn.KEY_TO_STN]
            >>> type(stn_loc_codes_dat)
            pandas.core.frame.DataFrame
            >>> stn_loc_codes_dat.head()
                  Station  ...                                    Former Operator
            0  Abbey Wood  ...  London & South Eastern Railway from 1 April 20...
            1  Abbey Wood  ...  London & South Eastern Railway from 1 April 20...
            2        Aber  ...  Keolis Amey Operations/Gweithrediadau Keolis A...
            3   Abercynon  ...  Keolis Amey Operations/Gweithrediadau Keolis A...
            4   Abercynon  ...  Keolis Amey Operations/Gweithrediadau Keolis A...
            [5 rows x 14 columns]
            >>> stn_loc_codes_dat.columns.to_list()
            ['Station',
             'Station Note',
             'ELR',
             'Mileage',
             'Status',
             'Degrees Longitude',
             'Degrees Latitude',
             'Grid Reference',
             'CRS',
             'CRS Note',
             'Owner',
             'Former Owner',
             'Operator',
             'Former Operator']
            >>> stn_loc_codes_dat[['Station', 'ELR', 'Mileage']].head()
                  Station  ELR   Mileage
            0  Abbey Wood  NKL  11m 43ch
            1  Abbey Wood  XRS  24.458km
            2        Aber  CAR   8m 69ch
            3   Abercynon  CAM  16m 28ch
            4   Abercynon  ABD  16m 28ch
        """

        if initial:
            args = {
                'data_name': validate_initial(initial),
                'method': self.collect_locations,
                'sub_dir': "a-z",
                'initial': initial,
            }
            kwargs.update(args)

            railway_station_data = self._fetch_data_from_file(
                update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        else:
            verbose_1 = collect_in_fetch_verbose(data_dir=dump_dir, verbose=verbose)
            verbose_2 = verbose_1 if is_home_connectable() else False

            data_sets = [
                self.fetch_locations(initial=x, update=update, verbose=verbose_2)
                for x in string.ascii_lowercase]

            if all(d[x] is None for d, x in zip(data_sets, string.ascii_uppercase)):
                if update:
                    print_inst_conn_err(verbose=verbose)
                    print_void_msg(data_name=self.KEY_TO_STN, verbose=verbose)

                data_sets = [
                    self.fetch_locations(x, update=False, verbose=verbose_1)
                    for x in string.ascii_lowercase
                ]

            stn_dat_tbl_ = (item[x] for item, x in zip(data_sets, string.ascii_uppercase))
            stn_dat_tbl = sorted(
                [x for x in stn_dat_tbl_ if x is not None], key=lambda x: x.shape[1], reverse=True)
            stn_data = pd.concat(stn_dat_tbl, axis=0, ignore_index=True, sort=False)

            stn_data = stn_data.where(pd.notna(stn_data), None)
            stn_data.sort_values(['Station'], inplace=True)

            stn_data.index = range(len(stn_data))

            last_updated_dates = (d[self.KEY_TO_LAST_UPDATED_DATE] for d in data_sets)
            latest_update_date = max(d for d in last_updated_dates if d is not None)

            railway_station_data = {
                self.KEY_TO_STN: stn_data,
                self.KEY_TO_LAST_UPDATED_DATE: latest_update_date,
            }

        if dump_dir is not None:
            self._save_data_to_file(
                data=railway_station_data, data_name=self.KEY_TO_STN, dump_dir=dump_dir,
                verbose=verbose)

        return railway_station_data



"""
Collects codes of `railway viaducts <http://www.railwaycodes.org.uk/tunnels/tunnels0.shtm>`_.
"""

import itertools
import re
import urllib.parse

from .._base import _Base
from ..parser import _get_last_updated_date, parse_table
from ..utils import home_page_url, is_home_connectable, print_inst_conn_err, print_void_msg, \
    validate_page_name


class Viaducts(_Base):
    """
    A class for collecting codes of
    `railway viaducts <http://www.railwaycodes.org.uk/tunnels/tunnels0.shtm>`_.
    """

    #: The name of the data.
    NAME: str = 'Railway viaducts'
    #: The key for accessing the data.
    KEY: str = 'Viaducts'

    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(home_page_url(), '/viaducts/viaducts0.shtm')

    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.other_assets import Viaducts  # from pyrcs import Viaducts
            >>> vdct = Viaducts()
            >>> vdct.NAME
            'Railway viaducts'
            >>> vdct.URL
            'http://www.railwaycodes.org.uk/viaducts/viaducts0.shtm'
        """

        super().__init__(
            data_dir=data_dir, content_type='catalogue', data_category="other-assets",
            update=update, verbose=verbose)

        self.page_range = range(1, 7)

    def _collect_codes(self, page_no, source, verbose=False):
        page_name = validate_page_name(self, page_no, valid_page_no=self.page_range)

        codes_dat, soup = parse_table(source=source, parser='html.parser', as_dataframe=True)

        codes_on_page = {
            page_name: codes_dat,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        data_name = re.sub(r"[()]", "", re.sub(r"[ -]", "-", page_name)).lower()
        self._save_data_to_file(data=codes_on_page, data_name=data_name, verbose=verbose)

        return codes_on_page

    def collect_codes(self, page_no, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects data of `railway viaducts`_ for a specific page number
        from the source web page.

        .. _`railway viaducts`: http://www.railwaycodes.org.uk/tunnels/tunnels0.shtm

        :param page_no: The page number to collect data from;
            valid values are ``1``, ``2``, ``3`` and ``4``.
        :type page_no: int | str
        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of railway viaducts for the specified ``page_no``
            and the date of when the data was last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import Viaducts  # from pyrcs import Viaducts
            >>> vdct = Viaducts()
            >>> page_1_codes = vdct.collect_codes(page_no=1)
            >>> type(page_1_codes)
            dict
            >>> list(page_1_codes.keys())
            ['Page 1 (A-C)', 'Last updated date']
            >>> page_1_dat = page_1_codes['Page 1 (A-C)']
            >>> type(page_1_dat)
            pandas.core.frame.DataFrame
            >>> page_1_dat.head()
                        Name  ... Spans
            0       7 Arches  ...     7
            1        36 Arch  ...    36
            2        42 Arch  ...
            3           A698  ...     5
            4  Abattoir Road  ...     8
            [5 rows x 7 columns]
        """

        data_name = self.NAME.lower()

        page_name = validate_page_name(self, page_no, valid_page_no=self.page_range)

        viaducts_codes = self._collect_data_from_source(
            data_name=data_name,
            method=self._collect_codes, initial=page_name, page_no=page_no,
            url=self.catalogue.get(page_name), confirmation_required=confirmation_required,
            confirmation_prompt=f"To collect data of {data_name} ({page_name})\n?",
            verbose=verbose, raise_error=raise_error)

        return viaducts_codes

    def fetch_codes(self, page_no=None, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `railway viaducts`_.

        .. _`railway viaducts`: http://www.railwaycodes.org.uk/tunnels/tunnels0.shtm

        :param page_no: The page number to collect data from;
            valid values are ``1``, ``2``, ``3`` and ``4``; defaults to ``None``.
        :type page_no: int | str
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of railway viaducts and
            the date of when the data was last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import Viaducts  # from pyrcs import Viaducts
            >>> vdct = Viaducts()
            >>> vdct_codes = vdct.fetch_codes()
            >>> type(vdct_codes)
            dict
            >>> list(vdct_codes.keys())
            ['Viaducts', 'Last updated date']
            >>> vdct.KEY
            'Viaducts'
            >>> vdct_codes_dat = vdct_codes[vdct.KEY]
            >>> type(vdct_codes_dat)
            dict
            >>> list(vdct_codes_dat.keys())
            ['Page 1 (A-C)',
             'Page 2 (D-G)',
             'Page 3 (H-K)',
             'Page 4 (L-P)',
             'Page 5 (Q-S)',
             'Page 6 (T-Z)']
            >>> page_6_codes = vdct_codes_dat['Page 6 (T-Z)']
            >>> type(page_6_codes)
            pandas.core.frame.DataFrame
            >>> page_6_codes.head()
                     Name                                  Notes  ... End mileage Spans
            0   Tadcaster  crosses River Wharfe; grade II listed  ...                11
            1        Taff                         see Red Bridge  ...
            2        Taff                                         ...
            3  Taff River                  also called Afon Taff  ...   170m 42ch
            4  Taffs Well                         see River Taff  ...
            [5 rows x 7 columns]
        """

        if page_no:
            page_name = validate_page_name(self, page_no, valid_page_no=self.page_range)

            args = {
                'data_name': re.sub(r"[()]", "", re.sub(r"[ -]", "-", page_name)).lower(),
                'method': self.collect_codes,
                'page_no': page_no,
            }
            kwargs.update(args)

            viaducts_codes = self._fetch_data_from_file(
                update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        else:
            verbose_1 = False if (dump_dir or not verbose) else (2 if verbose == 2 else True)
            verbose_2 = verbose_1 if is_home_connectable() else False

            codes_on_pages = [
                self.fetch_codes(page_no=page_no, update=update, verbose=verbose_2)
                for page_no in self.page_range]

            if all(x is None for x in codes_on_pages):
                if update:
                    print_inst_conn_err(verbose=verbose)
                    print_void_msg(data_name=self.KEY, verbose=verbose)

                codes_on_pages = [
                    self.fetch_codes(page_no=page_no, update=False, verbose=verbose_1)
                    for page_no in self.page_range]

            viaducts_codes = {
                self.KEY: {next(iter(x)): next(iter(x.values())) for x in codes_on_pages},
                self.KEY_TO_LAST_UPDATED_DATE:
                    max(next(itertools.islice(iter(x.values()), 1, 2)) for x in codes_on_pages),
            }

        if dump_dir is not None:
            self._save_data_to_file(
                data=viaducts_codes, data_name=self.KEY, dump_dir=dump_dir, verbose=verbose)

        return viaducts_codes



"""
A subpackage for collecting codes of
`other assets <http://www.railwaycodes.org.uk/otherassetsmenu.shtm>`_.

(See also the :py:class:`~pyrcs.collector.OtherAssets` class.)
"""

from .buzzer import Buzzer
from .depot import Depots
from .feature import Features
from .habd_wild import HabdWild
from .sig_box import SignalBoxes
from .station import Stations
from .telegraph import Telegraph
from .trough import WaterTroughs
from .tunnel import Tunnels
from .viaduct import Viaducts

__all__ = [
    'buzzer', 'Buzzer',
    'depot', 'Depots',
    'feature', 'Features',
    'habd_wild', 'HabdWild',
    'sig_box', 'SignalBoxes',
    'station', 'Stations',
    'telegraph', 'Telegraph',
    'trough', 'WaterTroughs',
    'tunnel', 'Tunnels',
    'viaduct', 'Viaducts',
]



"""
Collects codes of `HABD and WILD`_.

..`HABD and WILD`: http://www.railwaycodes.org.uk/misc/habdwild.shtm
"""

import urllib.parse

import bs4

from .._base import _Base
from ..parser import _get_last_updated_date, parse_tr
from ..utils import home_page_url


class HabdWild(_Base):
    """
    A class for collecting data of
    `HABDs and WILDs <http://www.railwaycodes.org.uk/features/habdwild.shtm>`_.

    .. note::

        - HABD: Hot axle box detector
        - WILD: Wheel impact load detector
    """

    #: The name of the data.
    NAME: str = 'Hot axle box detectors (HABDs) and wheel impact load detectors (WILDs)'
    #: The key for accessing the data.
    KEY: str = 'HABD and WILD'
    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(home_page_url(), '/features/habdwild.shtm')
    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.other_assets import HabdWild  # from pyrcs import HABDWILD
            >>> hw = HabdWild()
            >>> hw.NAME
            'Hot axle box detectors (HABDs) and wheel impact load detectors (WILDs)'
        """

        super().__init__(
            data_dir=data_dir, data_category="other-assets", update=update, verbose=verbose)

    def _collect_codes(self, source, verbose=False):
        try:
            sub_keys = self.KEY.split(' and ')
        except ValueError:
            sub_keys = [self.KEY + ' 1', self.KEY + ' 2']

        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        codes_list = []
        for h3 in soup.find_all('h3'):
            ths = [th.text.strip() for th in h3.find_next('thead').find_all('th')]
            trs = h3.find_next('tbody').find_all('tr')

            dat = parse_tr(trs=trs, ths=ths, as_dataframe=True)

            codes_list.append(dat)

        habds_and_wilds_codes_dat = dict(zip(sub_keys, codes_list))

        habds_and_wilds_codes = {
            self.KEY: habds_and_wilds_codes_dat,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=habds_and_wilds_codes, data_name=self.KEY, dump_dir=self._cdd("..", "features"),
            verbose=verbose)

        return habds_and_wilds_codes

    def collect_codes(self, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects codes of `HABDs and WILDs`_ from the source web page.

        .. _`HABDs and WILDs`: http://www.railwaycodes.org.uk/misc/habdwild.shtm

        .. note::

            - HABDs: Hot axle box detectors
            - WILDs: Wheel impact load detectors

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the codes of HABDs and WILDs and
            the date they were last updated, or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.other_assets import HabdWild  # from pyrcs import HABDWILD
            >>> hw = HabdWild()
            >>> hw_codes = hw.collect_codes()
            To collect data of HABD and WILD
            ? [No]|Yes: yes
            >>> type(hw_codes)
            dict
            >>> list(hw_codes.keys())
            ['HABD and WILD', 'Last updated date']
            >>> hw.KEY
            'HABD and WILD'
            >>> hw_codes_dat = hw_codes[hw.KEY]
            >>> type(hw_codes_dat)
            dict
            >>> list(hw_codes_dat.keys())
            ['HABD', 'WILD']
            >>> habd_dat = hw_codes_dat['HABD']
            >>> type(habd_dat)
            pandas.core.frame.DataFrame
            >>> habd_dat.head()
                ELR  ...                                              Notes
            0  BAG2  ...
            1  BAG2  ...  installed 29 September 1997, later moved to 74...
            2  BAG2  ...                             previously at 74m 51ch
            3  BAG2  ...                          removed 29 September 1997
            4  BAG2  ...           present in 1969, later moved to 89m 00ch
            [5 rows x 5 columns]
            >>> wild_dat = hw_codes_dat['WILD']
            >>> type(wild_dat)
            pandas.core.frame.DataFrame
            >>> wild_dat.head()
                ELR  ...                                              Notes
            0  AYR3  ...
            1  BAG2  ...
            2  BML1  ...
            3  BML1  ...
            4  CGJ3  ...  moved to 183m 68ch from 8 September 2018 / mov...
            [5 rows x 5 columns]
        """

        habds_and_wilds_codes = self._collect_data_from_source(
            data_name=self.KEY, method=self._collect_codes, url=self.URL,
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return habds_and_wilds_codes

    def fetch_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches codes of `HABDs and WILDs`_.

        .. _`HABDs and WILDs`: http://www.railwaycodes.org.uk/misc/habdwild.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the codes of HABDs and WILDs and
            the date they were last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.other_assets import HabdWild  # from pyrcs import HABDWILD
            >>> hw = HabdWild()
            >>> hw_codes = hw.fetch_codes()
            >>> type(hw_codes)
            dict
            >>> list(hw_codes.keys())
            ['HABD and WILD', 'Last updated date']
            >>> hw.KEY
            'HABD and WILD'
            >>> hw_codes_dat = hw_codes[hw.KEY]
            >>> type(hw_codes_dat)
            dict
            >>> list(hw_codes_dat.keys())
            ['HABD', 'WILD']
            >>> habd_dat = hw_codes_dat['HABD']
            >>> type(habd_dat)
            pandas.core.frame.DataFrame
            >>> habd_dat.head()
                ELR  ...                                              Notes
            0  BAG2  ...
            1  BAG2  ...  installed 29 September 1997, later moved to 74...
            2  BAG2  ...                             previously at 74m 51ch
            3  BAG2  ...                          removed 29 September 1997
            4  BAG2  ...           present in 1969, later moved to 89m 00ch
            [5 rows x 5 columns]
            >>> wild_dat = hw_codes_dat['WILD']
            >>> type(wild_dat)
            pandas.core.frame.DataFrame
            >>> wild_dat.head()
                ELR  ...                                              Notes
            0  AYR3  ...
            1  BAG2  ...
            2  BML1  ...
            3  BML1  ...
            4  CGJ3  ...  moved to 183m 68ch from 8 September 2018 / mov...
            [5 rows x 5 columns]
        """

        args = {
            'data_name': self.KEY,
            'method': self.collect_codes,
            'data_dir': self._cdd("..", "features"),
        }
        kwargs.update(args)

        habds_and_wilds_codes = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return habds_and_wilds_codes



"""
Collects data of
`CRS, NLC, TIPLOC and STANOX codes <http://www.railwaycodes.org.uk/crs/crs0.shtm>`_.
"""

import collections
import re
import string
import urllib.parse

import bs4
import pandas as pd
import requests
from pyhelpers._cache import _print_failure_message
from pyhelpers.dirs import validate_dir
from pyhelpers.ops import fake_requests_headers

from .._base import _Base
from ..parser import _get_last_updated_date, get_page_catalogue, parse_tr
from ..utils import collect_in_fetch_verbose, format_confirmation_prompt, home_page_url, \
    is_home_connectable, print_inst_conn_err, print_void_msg, validate_initial


def _parse_raw_location_name(x):
    """
    Parses the location name and extract any associated note from the raw data.

    :param x: Location name (in raw data).
    :type x: str | None
    :return: Location name and note (if any).
    :rtype: tuple

    **Examples**::

        >>> from pyrcs.line_data.loc_id import _parse_raw_location_name
        >>> _parse_raw_location_name(None)
        ('', '')
        >>> _parse_raw_location_name('Abbey Wood')
        ('Abbey Wood', '')
        >>> _parse_raw_location_name('Abercynon (formerly Abercynon South)')
        ('Abercynon', 'formerly Abercynon South')
        >>> _parse_raw_location_name('Allerton (reopened as Liverpool South Parkway)')
        ('Allerton', 'reopened as Liverpool South Parkway')
        >>> _parse_raw_location_name('Ashford International [domestic portion]')
        ('Ashford International', 'domestic portion')
        >>> _parse_raw_location_name('Ayr [unknown feature]')
        ('Ayr', 'unknown feature')
        >>> _parse_raw_location_name('Birkenhead Hamilton Square [see Hamilton Square]')
        ('Birkenhead Hamilton Square', 'see Hamilton Square')
    """

    if not x:
        x_, note = '', ''

    else:
        # Location name
        d = re.search(r'.*(?= \[[\"\']\()', x)
        if d is not None:
            x_ = d.group(0)
        elif ' [unknown feature' in x:  # ' [unknown feature, labelled "do not use"]' in x
            x_ = re.search(r'\w.*(?= \[unknown feature(, )?)', x).group(0)
        elif ') [formerly' in x:
            x_ = re.search(r'.*(?= \[formerly)', x).group(0)
        elif '✖' in x:
            x_ = re.search(r'.*(?=✖)', x).group(0)
        else:
            x_tmp = re.search(r'(?=[\[(]).*(?<=[])])|(?=\().*(?<=\) \[)', x)
            if x_tmp is not None:
                x_tmp = x_tmp.group(0)
                x_pat = re.compile(r'[Oo]riginally |'
                                   r'[Ff]ormerly |'
                                   r'[Ll]ater |'
                                   r'[Pp]resumed |'
                                   r' \(was |'
                                   r' \(in |'
                                   r' \(at |'
                                   r' \(also |'
                                   r' \(second code |'
                                   r'\?|'
                                   r'\n|'
                                   r' \(\[\'|'
                                   r' \(definition unknown\)|'
                                   r' \(reopened |'
                                   r'( portion])$|'
                                   r'[Ss]ee ')
                x_ = ' '.join(x.replace(x_tmp, '').split()) if re.search(x_pat, x) else x
            else:
                x_ = x

        # Note
        y_ = x.replace(x_, '', 1).strip()
        if y_ == '':
            note = ''
        elif '✖' in y_:
            note = re.search(r'(?<=✖).*', y_).group(0)
        else:
            note_ = re.search(r'(?<=[\[(])[\w ,?]+(?=[])])', y_)
            if note_ is None:
                note_ = re.search(
                    r'(?<=(\[[\'\"]\()|(\([\'\"]\[)|(\) \[)).*(?=(\)[\'\"]])|(][\'\"]\))|])',
                    y_)
            elif '"now deleted"' in y_ and y_.startswith('(') and y_.endswith(')'):
                note_ = re.search(r'(?<=\().*(?=\))', y_)

            note = note_.group(0) if note_ is not None else ''
            if note.endswith('\'') or note.endswith('"'):
                note = note[:-1]

        if 'STANOX ' in x_ and 'STANOX ' in x and note == '':
            x_ = x[0:x.find('STANOX')].strip()
            note = x[x.find('STANOX'):]

    return x_, note


def _amendment_to_location_names():
    """
    Creates a dictionary of amendments to adjust location names using regular expressions.

    This method generates a dictionary where the keys are regular expression patterns, and the
    values are the corresponding amendments to be applied to location names.

    :return: A dictionary where each key is a regular expression pattern and each value is the
        corresponding replacement string to amend the location names.
    :rtype: dict

    **Examples**::

        >>> from pyrcs.line_data.loc_id import _amendment_to_location_names
        >>> loc_name_amendment_dict = _amendment_to_location_names()
        >>> list(loc_name_amendment_dict.keys())
        ['Location']
    """

    location_name_amendment_dict = {
        'Location': {re.compile(r' And | \+ '): ' & ',
                     re.compile(r'-By-'): '-by-',
                     re.compile(r'-In-'): '-in-',
                     re.compile(r'-En-Le-'): '-en-le-',
                     re.compile(r'-La-'): '-la-',
                     re.compile(r'-Le-'): '-le-',
                     re.compile(r'-On-'): '-on-',
                     re.compile(r'-The-'): '-the-',
                     re.compile(r' Of '): ' of ',
                     re.compile(r'-Super-'): '-super-',
                     re.compile(r'-Upon-'): '-upon-',
                     re.compile(r'-Under-'): '-under-',
                     re.compile(r'-Y-'): '-y-'}}

    return location_name_amendment_dict


def _extra_annotations():
    extra_annotations = [
        ('✖Earlier code ', '✖Later code'),
        ('✖Later code ', '✖Earlier code'),
        ('✖Earlier code, soon replaced ', '✖Later code introduced before station opened'),
        ('✖Later code introduced before station opened ', '✖Earlier code, soon replaced'),
        ('✖Original code ', '✖Later code'),
        ('✖Later code ', '✖Original code'),
        ('✖Original code ', '✖Later code from station opening'),
        ('✖Later code from station opening ', '✖Original code'),
        ('✖Newer designation? ', '✖Older designation?'),
        ('✖Older designation? ', '✖Newer designation?'),
        ('✖Both codes quoted with equal reliability ',
         '✖Both codes quoted with equal reliability'),
        ('✖Code used by operational research software ',
         '✖Code as used by National Rail Enquiries and other public systems'),
        ('✖Code as used by National Rail Enquiries and other public systems ',
         '✖Code used by operational research software'),
        ('✖Code assigned in error ', '✖Corrected code'),
        ('✖Original code ', '✖Later code after becoming part of national network'),
        ('✖Later code after becoming part of national network ', '✖Original code'),
        ('✖Code used after station reopened ', '✖Code used until station reopened'),
        ('✖Code used until station reopened ', '✖Code used after station reopened'),
        ('✖Code used after nearby station reopened ', '✖Code used until nearby station reopened'),
        ('✖Code used until nearby station reopened ', '✖Code used after nearby station reopened'),
        ('✖Original code; see also CRS explanation ', '✖Later code; see also CRS explanation'),
        ('✖Later code; see also CRS explanation ', '✖Original code; see also CRS explanation'),
        ('✖Most sources state LOUNOC, some have LOUNGOC ',
         '✖Most sources state LOUNOC, some have LOUNGOC'),
        ("✖Code may start with 'R' ", "✖Code may start with 'B'"),
        ("✖Code may start with 'B' ", "✖Code may start with 'R'"),
        ('✖Older designation when station open ', '✖Newer designation once station closed'),
        ('✖Newer designation once station closed ', '✖Older designation when station open'),
        ('✖Original code when named Butlins Penychain ', '✖Later code after renaming Penychain'),
        ('✖Later code after renaming Penychain ', '✖Original code when named Butlins Penychain'),
        ('✖Earlier code ',
         '✖Later code after East London Line operated as part of national network'),
        ('✖Code should be ROODENDGD but some listings show this as RO0DENDGD (with zero digit) ',
         '✖Code should be ROODENDGD but some listings show this as RO0DENDGD (with zero digit)'),
        ('✖Earlier code? ', '✖Later code'),
        ('✖See CRS explanation ', '✖Code believed listed in error'),
        ('✖Code is 56582 but some sources have 56782 ',
         '✖Code is 56582 but some sources have 56782'),
        ('✖Code should be 79251 but one source has 79521 ',
         '✖Code should be 79251 but one source has 79521'),
        ('✖Code is 56540 but some sources have 56450 ',
         '✖Code is 56540 but some sources have 56450'),
        ('✖Earlier code ', '✖Later code; also see CRS explanation'),
        ('✖Later code ', '✖Possibly earlier code'),
        ('✖Code is 142505 but is also reported as 145505 ',
         '✖Code is 142505 but is also reported as 145505'),
        ('✖Earlier code ', '✖Later code after station closed'),
    ]

    return extra_annotations


def _count_sep(x):
    if '\r\n' in x:
        r_n_counts = x.count('\r\n')

    elif '\r' in x:
        r_n_counts = x.count('\r')

    else:  # Ad hoc
        if '~LO\n' in x:
            x = x.replace('~LO\n', '')
        # elif any(all(a_ in x for a_ in a) for a in self._extra_annotations()):
        #     temp = [
        #         x.replace(a[0], f'{a[0][:-1]}\n') for a in self._extra_annotations()
        #         if a[0] in x and x.endswith(a[1])]
        #     x = temp[0]
        r_n_counts = x.count('\n')

    return r_n_counts


def _split_dat_and_note(x):
    if '\r\n' in x:
        x_ = x.split('\r\n')

    elif '\r' in x:
        x_ = x.split('\r')

    elif '\n' in x:
        if '~LO\n' in x:  # Ad hoc
            x = x.replace('~LO\n', '')
        x_ = x.split('\n')

    else:
        x_ = x

    return x_


def _fix_exceptional_cases(data):
    x1 = 'Ely Papworth Sidings English Welsh & Scottish Railway International\n' \
         'Ely Papworth Sidings DB Schenker International'

    if x1 in data['Location'].values:
        x1_ = 'Ely Papworth Sidings English Welsh & Scottish Railway International\n' \
              'Ely Papworth Sidings DB Schenker International\n' \
              'Ely Papworth Sidings DB Schenker International'

        data.loc[data['Location'] == x1, 'Location'] = x1_

    return data


def _parse_mult_alt_codes(data):
    """
    Cleanses multiple alternatives for every code column.

    :param data: The preprocessed data of the location codes.
    :type data: pandas.DataFrame
    :return: The cleansed data of the location codes where multiple alternatives are replicated.
    :rtype: pandas.DataFrame
    """

    data_ = data.copy()
    data_ = _fix_exceptional_cases(data_)

    code_col_names = ['Location', 'CRS', 'NLC', 'TIPLOC', 'STANME', 'STANOX']

    r_n_counts = data_[code_col_names].map(_count_sep)
    # # Debugging:
    # for col in code_col_names:
    #     for i, x in enumerate(data_[col]):
    #         try:
    #             lid._count_sep(x)
    #         except Exception:
    #             print(col, i, x)
    #             break
    r_n_counts_ = r_n_counts.mul(-1).add(r_n_counts.max(axis=1), axis='index')

    for col in code_col_names:
        for i in data_.index:
            d = r_n_counts_.loc[i, col]
            x = data_.loc[i, col]
            if d > 0:
                if '\r\n' in x:
                    if col == 'Location':
                        data_.loc[i, col] = x + ''.join(['\r\n' + x.split('\r\n')[-1]] * d)
                    else:
                        data_.loc[i, col] = x + ''.join(['\r\n'] * d)
                elif '\r' in x:
                    if col == 'Location':
                        data_.loc[i, col] = x + ''.join(['\r' + x.split('\r')[-1]] * d)
                    else:
                        data_.loc[i, col] = x + ''.join(['\r'] * d)
                else:  # e.g. '\n' in dat:
                    if col == 'Location':
                        data_.loc[i, col] = '\n'.join([x] * (d + 1))
                    else:
                        data_.loc[i, col] = x + ''.join(['\n'] * d)
            # elif any(all(a_ in x for a_ in a) for a in self._extra_annotations()):
            #     temp = [
            #         x.replace(a[0], f'{a[0][:-1]}\n') for a in self._extra_annotations()
            #         if a[0] in x and x.endswith(a[1])]
            #     data_.loc[i, col] = temp[0]

    data_[code_col_names] = data_[code_col_names].map(_split_dat_and_note)

    data_ = data_.explode(code_col_names, ignore_index=True)

    temp = data_.select_dtypes(['object'])
    data_[temp.columns] = temp.apply(lambda x_: x_.str.strip())

    return data_


def _parse_code_note(x):
    """
    Gets note for every code column.

    :param x: The raw data of a given code.
    :type x: str | None
    :return: The extra information (if any) about the given code.
    :rtype: str

    **Examples**::

        >>> from pyrcs.line_data.loc_id import _parse_code_note
        >>> _parse_code_note('860260✖Earlier code')
        ('860260', 'Earlier code')
    """

    if not x:
        y, note = x, ''

    else:
        if '✖' in x:
            y, note = map(str.strip, x.split('✖', 1))

        else:  # Search for notes
            n1 = re.search(r'(?<=[\[(])[\w,? ]+(?=[)\]])', x)

            if n1:
                note = n1.group(0)
                y = x.replace(note, '').strip('[(\')] ')

                n2 = re.search(r'[\w ,]+(?= [\[(\'])', note)  # Remove redundant characters
                if n2:
                    note = n2.group(0)

            else:
                y, note = x.strip(), ''  # Default case: No note found

    return y, note


def _parse_code_notes(data):
    """
    Gets notes for every code column.

    :param data: The preprocessed data of the location codes.
    :type data: pandas.DataFrame
    """

    codes_col_names = ['CRS', 'NLC', 'TIPLOC', 'STANME', 'STANOX']

    for col in codes_col_names:
        data[[col, col + '_Note']] = pd.DataFrame(
            data[col].map(_parse_code_note).to_list(), index=data.index)
    # # Debugging:
    # for col in codes_col_names:
    #     for i, x in enumerate(data[col]):
    #         try:
    #             lid._get_code_note(x)
    #         except Exception:
    #             print(col, i, x)
    #             break

    return data


def _stanox_note(x):  # Parse STANOX note
    """
    Parses STANOX note.

    :param x: STANOX note.
    :type x: str | None
    :return: STANOX and its corresponding note.
    :rtype: tuple
    """

    if x in ('-', '') or x is None:
        stanox, note = '', ''

    else:
        if re.match(r'\d{5}$', x):
            stanox = x
            note = ''

        elif re.match(r'\d{5}\*$', x):
            stanox = x.rstrip('*')
            note = 'Pseudo STANOX'

        elif re.match(r'\d{5} \w.*', x):
            stanox = re.search(r'\d{5}', x).group()
            note = re.search(r'(?<= )\w.*', x).group()

        else:
            d = re.search(r'[\w *,]+(?= [\[(\'])', x)
            stanox = d.group() if d is not None else x
            note = 'Pseudo STANOX' if '*' in stanox else ''
            n = re.search(r'(?<=[\[(\'])[\w, ]+.(?=[)\]\'])', x)

            if n is not None:
                note = '; '.join(x for x in [note, n.group()] if x != '')

            if '(' not in note and note.endswith(')'):
                note = note.rstrip(')')

    return stanox, note


def _parse_stanox_note(data):
    """
    Parses the note for STANOX.

    :param data: The preprocessed data of the location codes.
    :type data: pandas.DataFrame
    """

    col_name = 'STANOX'
    note_col_name = col_name + '_Note'

    # noinspection SpellCheckingInspection
    data[col_name] = data[col_name].str.replace('NANAN', '')

    if not data.empty:
        parsed_dat = data[col_name].map(_stanox_note).to_list()
        data[[col_name, note_col_name]] = pd.DataFrame(parsed_dat, index=data.index)
    else:
        # No data is available on the web page for the given 'key_word'
        data[note_col_name] = data[col_name]

    data[col_name] = data[col_name].str.replace('-', '')

    return data


class LocationIdentifiers(_Base):
    """
    A class for collecting data of location identifiers
    (including `CRS, NLC, TIPLOC and STANOX codes <http://www.railwaycodes.org.uk/crs/crs0.shtm>`_
    and `other systems' station codes <http://www.railwaycodes.org.uk/crs/crs1.shtm>`_).

    The class retrieves and organises location identifiers for various railway stations and
    other related systems.
    """

    #: The name of the data.
    NAME: str = 'CRS, NLC, TIPLOC and STANOX codes'
    #: The key for accessing the data.
    KEY: str = 'Location ID'

    #: The URL of the main web page for the data.
    URL: str = urllib.parse.urljoin(home_page_url(), '/crs/crs0.shtm')

    #: The key for accessing the data of *other systems*.
    KEY_TO_OTHER_SYSTEMS: str = 'Other systems'
    #: The key for accessing the data of *multiple station codes explanatory note*.
    KEY_TO_MSCEN: str = 'Multiple station codes explanatory note'
    #: The key for accessing the data of *additional notes*.
    KEY_TO_NOTES: str = 'Notes'

    #: The key used to reference the last updated date in the data.
    KEY_TO_LAST_UPDATED_DATE: str = 'Last updated date'

    def __init__(self, data_dir=None, update=False, verbose=True):
        """
        :param data_dir: The name of the directory for storing the data; defaults to ``None``.
        :type data_dir: str | None
        :param update: Whether to check for updates to the data catalogue; defaults to ``False``.
        :type update: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int

        :ivar dict catalogue: The catalogue of the data.
        :ivar str last_updated_date: The date when the data was last updated.
        :ivar str data_dir: The path to the directory containing the data.
        :ivar str current_data_dir: The path to the current data directory.

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> lid.NAME
            'CRS, NLC, TIPLOC and STANOX codes'
            >>> lid.URL
            'http://www.railwaycodes.org.uk/crs/crs0.shtm'
        """

        # data_cluster = re.sub(r",| codes| and", "", self.NAME.lower()).replace(" ", "-")
        super().__init__(
            data_dir=data_dir, content_type='catalogue', data_category="line-data",
            data_cluster="crs-nlc-tiploc-stanox", update=update, verbose=verbose)

        # Adds the multiple station codes explanatory note (MSCEN) to the catalogue
        mscen_url = urllib.parse.urljoin(home_page_url(), '/crs/crs2.shtm')
        self.catalogue.update({self.KEY_TO_MSCEN: mscen_url})

        # Retrieve the catalogue for other systems' station codes
        other_systems_url = self.catalogue[self.KEY_TO_OTHER_SYSTEMS]
        self.other_systems_catalogue = get_page_catalogue(url=other_systems_url)

    @staticmethod
    def _parse_notes_page(source, parser='html.parser'):
        """
        Parses the additional note page at the specified URL and extract its contents.

        :param parser: The `parser`_ to use with `bs4.BeautifulSoup`_
            (e.g. ``'html.parser'``, ``'lxml'``); defaults to ``'html.parser'``.
        :type parser: str
        :return: A list of parsed text elements from the web page.
        :rtype: list

        .. _`parser`:
            https://www.crummy.com/software/BeautifulSoup/bs4/doc/
            index.html#specifying-the-parser-to-use
        .. _`bs4.BeautifulSoup`:
            https://www.crummy.com/software/BeautifulSoup/bs4/doc/index.html

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> url = 'http://www.railwaycodes.org.uk/crs/crs2.shtm'
            >>> parsed_note_dat = lid._parse_notes_page(note_url=url)
            >>> parsed_note_dat[3]
                               Location  CRS CRS_alt1 CRS_alt2
            0           Glasgow Central  GLC      GCL
            1      Glasgow Queen Street  GLQ      GQL
            2                   Heworth  HEW      HEZ
            3      Highbury & Islington  HHY      HII      XHZ
            4    Lichfield Trent Valley  LTV      LIF
            5     Liverpool Lime Street  LIV      LVL
            6   Liverpool South Parkway  LPY      ALE
            7         London St Pancras  STP      SPL      SPX
            8                   Retford  RET      XRO
            9   Smethwick Galton Bridge  SGB      GTI
            10                 Tamworth  TAM      TAH
            11       Willesden Junction  WIJ      WJH      WJL
            12   Worcestershire Parkway  WOP      WPH
        """

        soup = bs4.BeautifulSoup(markup=source.content, features=parser)
        contents = soup.find_all(['p', 'pre'])

        raw_text = filter(
            None, [x.get_text(strip=True) for x in contents if isinstance(x.next_element, str)])

        notes = []
        for x in raw_text:
            if '\n' in x and '\t' in x:
                text = re.sub('\t+', ',', x).replace('\t', ' ').replace('\xa0', '').split('\n')
            else:
                text = x.replace('\t', ' ').replace('\xa0', '')

            if isinstance(text, list):
                text = [[x.strip() for x in t.split(',')] for t in text if t != '']
                text = [x + [''] if len(x) < 4 else x for x in text]
                temp = pd.DataFrame(text, columns=['Location', 'CRS', 'CRS_alt1', 'CRS_alt2'])
                notes.append(temp.fillna(''))
            else:
                to_remove = ['click the link', 'click your browser', 'Thank you', 'shown below']
                if text.strip() != '' and not any(t in text for t in to_remove):
                    notes.append(text)

        return notes, soup

    # -- CRS, NLC, TIPLOC and STANOX ---------------------------------------------------------------

    def _parse_crs_notes(self, data, initial, soup):
        if any('see note' in crs_note for crs_note in data['CRS_Note']):
            indices = [i for i, crs_n in enumerate(data['CRS_Note']) if 'see note' in crs_n]

            notes = []
            for x in soup.find_all('a', href=True, string='note'):
                source = requests.get(
                    url=urllib.parse.urljoin(self.catalogue[initial], x['href']),
                    headers=fake_requests_headers())

                notes.append(self._parse_notes_page(source)[0])

            loc_id_notes = dict(zip(data['CRS'].iloc[indices], notes))

        else:
            loc_id_notes = None

        return loc_id_notes

    def _collect_loc_id(self, initial, source, verbose=False):
        initial_ = validate_initial(x=initial)

        # url = lid.catalogue[initial_]
        # source = requests.get(url)
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        thead, tbody = soup.find('thead'), soup.find('tbody')
        ths, trs = [th.get_text(strip=True) for th in thead.find_all('th')], tbody.find_all('tr')

        dat = parse_tr(trs=trs, ths=ths, sep=None, as_dataframe=True)
        dat = dat.replace({'\b-\b': '', '\xa0\xa0': ' ', '&half;': ' and 1/2'}, regex=True)

        data = dat.replace({'\xa0': ''}, regex=True)

        # Parse location names and their corresponding notes
        data[['Location', 'Location_Note']] = pd.DataFrame(  # Collect additional info as note
            data['Location'].map(_parse_raw_location_name).to_list())
        # # Debugging
        # for i, x in enumerate(data['Location']):
        #     try:
        #         _parse_location_name(x)
        #     except Exception:
        #         print(i)
        #         break
        data.replace(_amendment_to_location_names(), regex=True, inplace=True)

        # Cleanse multiple alternatives for every code column
        data = _parse_mult_alt_codes(data=data)

        # Parse note for every code column
        data = _parse_code_notes(data=data)

        # Parse STANOX note
        data = _parse_stanox_note(data=data)

        loc_codes = {
            initial_: data,
            self.KEY_TO_NOTES: self._parse_crs_notes(data=data, initial=initial_, soup=soup),
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(data=loc_codes, data_name=initial_, sub_dir="a-z", verbose=verbose)

        return loc_codes

    def collect_loc_id(self, initial, confirmation_required=True, verbose=False, raise_error=False):
        """
        Collects `CRS, NLC, TIPLOC, STANME and STANOX codes
        <http://www.railwaycodes.org.uk/crs/crs0.shtm>`_ for a given initial letter.

        :param initial: The initial letter (e.g. ``'a'``, ``'z'``) of a location name.
        :type initial: str
        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing data of locations whose names start with the given
            initial letter, along with the date of the last update.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> loc_a_codes = lid.collect_loc_id(initial='a')
            To collect data of CRS, NLC, TIPLOC and STANOX codes beginning with "A"
            ? [No]|Yes: yes
            >>> type(loc_a_codes)
            dict
            >>> list(loc_a_codes.keys())
            ['A', 'Additional notes', 'Last updated date']
            >>> loc_a_codes_dat = loc_a_codes['A']
            >>> type(loc_a_codes_dat)
            pandas.core.frame.DataFrame
            >>> loc_a_codes_dat.head()
                                          Location CRS  ... STANME_Note STANOX_Note
            0                 1999 Reorganisations      ...
            1                                   A1      ...
            2                       A463 Traded In      ...
            3  A483 Road Scheme Supervisors Closed      ...
            4                               Aachen      ...
            [5 rows x 12 columns]
        """

        initial_ = validate_initial(x=initial)

        loc_id_data = self._collect_data_from_source(
            data_name=self.NAME, method=self._collect_loc_id, initial=initial_,
            additional_fields=self.KEY_TO_NOTES, confirmation_required=confirmation_required,
            verbose=verbose, raise_error=raise_error)

        return loc_id_data

    def fetch_loc_id(self, initial=None, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `CRS, NLC, TIPLOC, STANME and STANOX codes`_.

        .. _`CRS, NLC, TIPLOC, STANME and STANOX codes`:
            http://www.railwaycodes.org.uk/crs/crs0.shtm

        :param initial: The initial letter (e.g. ``'a'``, ``'z'``) of a location name.
        :type initial: str
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: Path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``True``.
        :type verbose: bool | int
        :return: A dictionary containing data of locations whose names start with the given
            initial letter, along with the date of the last update.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> loc_a_codes = lid.fetch_loc_id(initial='a')
            >>> type(loc_a_codes)
            dict
            >>> list(loc_a_codes.keys())
            ['A', 'Additional notes', 'Last updated date']
            >>> loc_a_codes_dat = loc_a_codes['A']
            >>> type(loc_a_codes_dat)
            pandas.core.frame.DataFrame
            >>> loc_a_codes_dat.head()
                                          Location CRS  ... STANME_Note STANOX_Note
            0                 1999 Reorganisations      ...
            1                                   A1      ...
            2                       A463 Traded In      ...
            3  A483 Road Scheme Supervisors Closed      ...
            4                               Aachen      ...
            [5 rows x 12 columns]
            >>> loc_codes = lid.fetch_loc_id()
            >>> list(loc_codes.keys())
            ['Location ID', 'Last updated date']
            >>> loc_codes[lid.KEY].shape
            (59873, 12)
        """

        if initial:
            args = {
                'data_name': validate_initial(initial),
                'method': self.collect_loc_id,
                'sub_dir': "a-z",
                'initial': initial,
            }
            kwargs.update(args)

            loc_id_data = self._fetch_data_from_file(
                update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        else:
            verbose_1 = collect_in_fetch_verbose(data_dir=dump_dir, verbose=verbose)

            # Get every data table
            verbose_2 = verbose_1 if is_home_connectable() else False
            dat_list = [
                self.fetch_loc_id(initial=x, update=update, verbose=verbose_2)
                for x in string.ascii_lowercase]

            if all(d[x] is None for d, x in zip(dat_list, string.ascii_uppercase)):
                if update:
                    print_inst_conn_err(verbose=verbose)
                    print_void_msg(data_name=self.KEY, verbose=verbose)

                dat_list = [
                    self.fetch_loc_id(initial=x, update=False, verbose=verbose_1)
                    for x in string.ascii_lowercase]

            # Select DataFrames only
            data = pd.concat(
                (item[x] for item, x in zip(dat_list, string.ascii_uppercase)), ignore_index=True)

            # Get the latest updated date
            last_updated_dates = (
                item[self.KEY_TO_LAST_UPDATED_DATE]
                for item, _ in zip(dat_list, string.ascii_uppercase))
            latest_update_date = max(d for d in last_updated_dates if d is not None)

            loc_id_data = {self.KEY: data, self.KEY_TO_LAST_UPDATED_DATE: latest_update_date}

        return loc_id_data

    # -- Other systems -----------------------------------------------------------------------------

    @staticmethod
    def _parse_code(x):
        protocol = 'https://'

        if '; ' in x and protocol in x:
            temp = x.split('; ')
            x0, x1 = temp[0], [y.split(protocol) for y in temp if protocol in y][0]
            x1 = ' ('.join([x1[0], protocol + x1[1]]) + ')'
            x_ = [x0, x1]
        else:
            x_ = [x, '']

        return x_

    def _parse_tbl_dat(self, h3_or_h4):
        tbl_dat = h3_or_h4.find_next('thead'), h3_or_h4.find_next('tbody')
        thead, tbody = tbl_dat
        ths = [x.text for x in thead.find_all('th')]
        trs = tbody.find_all('tr')
        tbl = parse_tr(trs=trs, ths=ths, as_dataframe=True)

        if 'Code' in tbl.columns:
            if tbl.Code.str.contains('https://').sum() > 0:
                temp = tbl['Code'].map(self._parse_code)
                tbl_ext = pd.DataFrame(zip(*temp)).T
                tbl_ext.columns = ['Code', 'Code_extra']
                del tbl['Code']
                tbl = pd.concat([tbl, tbl_ext], axis=1, sort=False)

        return tbl

    def _collect_other_systems_codes(self, source, verbose=False):
        soup = bs4.BeautifulSoup(markup=source.content, features='html.parser')

        other_systems_codes = collections.defaultdict(dict)

        for h3 in soup.find_all('h3'):
            h4 = h3.find_next('h4')

            if h4 is not None:
                while h4:
                    prev_h3 = h4.find_previous('h3')
                    if prev_h3.text == h3.text:
                        other_systems_codes[h3.text].update(
                            {h4.text: self._parse_tbl_dat(h4)})
                        h4 = h4.find_next('h4')
                    elif h3.text not in other_systems_codes.keys():
                        other_systems_codes.update({h3.text: self._parse_tbl_dat(h3)})
                        break
                    else:
                        break
            else:
                other_systems_codes.update({h3.text: self._parse_tbl_dat(h3)})

        other_systems_codes = {
            self.KEY_TO_OTHER_SYSTEMS: other_systems_codes,
            self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup),
        }

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(
            data=other_systems_codes, data_name=self.KEY_TO_OTHER_SYSTEMS, verbose=verbose)

        return other_systems_codes

    def collect_other_systems_codes(self, confirmation_required=True, verbose=False,
                                    raise_error=False):
        """
        Collects data of `other systems' station codes`_ from the source web page.

        .. _`other systems' station codes`: http://www.railwaycodes.org.uk/crs/crs1.shtm

        :param confirmation_required: Whether user confirmation is required before proceeding;
            defaults to ``True``.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception; defaults to ``True``.
            if ``raise_error=False``, the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing station codes for other systems,
            or ``None`` if no data is collected.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> os_codes = lid.collect_other_systems_codes()
            To collect data of Other systems
            ? [No]|Yes: yes
            >>> type(os_codes)
            dict
            >>> list(os_codes.keys())
            ['Other systems', 'Last updated date']
            >>> lid.KEY_TO_OTHER_SYSTEMS
            'Other systems'
            >>> os_codes_dat = os_codes[lid.KEY_TO_OTHER_SYSTEMS]
            >>> type(os_codes_dat)
            collections.defaultdict
            >>> list(os_codes_dat.keys())
            ['Córas Iompair Éireann (Republic of Ireland)',
             'Crossrail',
             'Croydon Tramlink',
             'Docklands Light Railway',
             'Manchester Metrolink',
             'Translink (Northern Ireland)',
             'Tyne & Wear Metro']
        """

        other_systems_codes = self._collect_data_from_source(
            data_name=self.KEY_TO_OTHER_SYSTEMS.lower(), method=self._collect_other_systems_codes,
            url=self.catalogue[self.KEY_TO_OTHER_SYSTEMS],
            confirmation_required=confirmation_required, verbose=verbose, raise_error=raise_error)

        return other_systems_codes

    def fetch_other_systems_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches data of `other systems' station codes`_.

        .. _`other systems' station codes`: http://www.railwaycodes.org.uk/crs/crs1.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing station codes for other systems.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> os_codes = lid.fetch_other_systems_codes()
            >>> type(os_codes)
            dict
            >>> list(os_codes.keys())
            ['Other systems', 'Last updated date']
            >>> lid.KEY_TO_OTHER_SYSTEMS
            'Other systems'
            >>> os_codes_dat = os_codes[lid.KEY_TO_OTHER_SYSTEMS]
            >>> type(os_codes_dat)
            collections.defaultdict
            >>> list(os_codes_dat.keys())
            ['Córas Iompair Éireann (Republic of Ireland)',
             'Crossrail',
             'Croydon Tramlink',
             'Docklands Light Railway',
             'Manchester Metrolink',
             'Translink (Northern Ireland)',
             'Tyne & Wear Metro']
        """

        kwargs.update(
            {'data_name': self.KEY_TO_OTHER_SYSTEMS, 'method': self.collect_other_systems_codes})

        other_systems_codes = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return other_systems_codes

    # -- Additional notes --------------------------------------------------------------------------

    def _collect_notes(self, source, verbose=False, parser='html.parser'):
        notes_dat, soup = self._parse_notes_page(source=source, parser=parser)

        notes = {}
        explanatory_notes, additional_notes = {}, []

        for x in notes_dat:
            if isinstance(x, str):
                if 'Last update' not in x:
                    additional_notes.append(x)
            else:
                explanatory_notes.update({self.KEY_TO_MSCEN: x})

        notes.update(
            {self.KEY_TO_NOTES: explanatory_notes | {'Additional notes': additional_notes},
             self.KEY_TO_LAST_UPDATED_DATE: _get_last_updated_date(soup=soup)}
        )

        if verbose in {True, 1}:
            print("Done.")

        self._save_data_to_file(data=notes, data_name=self.KEY_TO_NOTES, verbose=verbose)

        return notes

    def collect_notes(self, confirmation_required=True, verbose=False, raise_error=False):
        # noinspection PyShadowingNames
        """
        Collects the explanatory note related to multiple station codes (CRS codes)
        from the source web page.

        :param confirmation_required: Whether user confirmation is required;
            if ``confirmation_required=True`` (default), prompts the user for confirmation
            before proceeding with data collection.
        :type confirmation_required: bool
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :param raise_error: Whether to raise the provided exception;
            if ``raise_error=False`` (default), the error will be suppressed.
        :type raise_error: bool
        :return: A dictionary containing the data of the multiple station codes explanatory note,
            or ``None`` if the note is not found or the collection is not performed.
        :rtype: dict | None

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> explanatory_notes = lid.collect_notes()
            To collect data of multiple station codes explanatory note
            ? [No]|Yes: yes
            >>> type(explanatory_notes)
            dict
            >>> list(explanatory_notes.keys())
            ['Multiple station codes explanatory note', 'Notes', 'Last updated date']
            >>> lid.KEY_TO_MSCEN
            'Multiple station codes explanatory note'
            >>> explanatory_notes_ = explanatory_notes[lid.KEY_TO_MSCEN]
            >>> type(explanatory_notes_)
            pandas.core.frame.DataFrame
            >>> explanatory_notes_.head()
                              Location  CRS CRS_alt1 CRS_alt2
            0  Ebbsfleet International  EBD      EBF
            1          Glasgow Central  GLC      GCL
            2     Glasgow Queen Street  GLQ      GQL
            3                  Heworth  HEW      HEZ
            4     Highbury & Islington  HHY      HII      XHZ
        """

        explanatory_notes = self._collect_data_from_source(
            data_name=self.KEY_TO_MSCEN.lower(), method=self._collect_notes,
            url=self.catalogue[self.KEY_TO_MSCEN],
            confirmation_required=confirmation_required,
            confirmation_prompt=format_confirmation_prompt(data_name=self.KEY_TO_MSCEN.lower()),
            verbose=verbose, raise_error=raise_error)

        return explanatory_notes

    def fetch_notes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches the explanatory note for multiple station codes (CRS codes).

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing the data of the multiple station codes explanatory note.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> exp_note = lid.fetch_notes()
            >>> type(exp_note)
            dict
            >>> list(exp_note.keys())
            ['Multiple station codes explanatory note', 'Notes', 'Last updated date']
            >>> lid.KEY_TO_MSCEN
            'Multiple station codes explanatory note'
            >>> exp_note_dat = exp_note[lid.KEY_TO_MSCEN]
            >>> type(exp_note_dat)
            pandas.core.frame.DataFrame
            >>> exp_note_dat.head()
                             Location  CRS CRS_alt1 CRS_alt2
            0         Glasgow Central  GLC      GCL
            1    Glasgow Queen Street  GLQ      GQL
            2                 Heworth  HEW      HEZ
            3    Highbury & Islington  HHY      HII      XHZ
            4  Lichfield Trent Valley  LTV      LIF
        """

        kwargs.update({'data_name': self.KEY_TO_NOTES, 'method': self.collect_notes})

        explanatory_note = self._fetch_data_from_file(
            update=update, dump_dir=dump_dir, verbose=verbose, **kwargs)

        return explanatory_note

    # -- All codes ---------------------------------------------------------------------------------

    def fetch_codes(self, update=False, dump_dir=None, verbose=False, **kwargs):
        """
        Fetches location codes listed in the `CRS, NLC, TIPLOC and STANOX codes`_ catalogue
        (including `other systems' station codes`_).

        .. _`CRS, NLC, TIPLOC and STANOX codes`:
            http://www.railwaycodes.org.uk/crs/crs0.shtm
        .. _`other systems' station codes`:
            http://www.railwaycodes.org.uk/crs/crs1.shtm

        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_dir: The path to a directory where the data file will be saved;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary containing location codes and date of when the data was last updated.
        :rtype: dict

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> loc_codes = lid.fetch_codes()
            >>> type(loc_codes)
            dict
            >>> list(loc_codes.keys())
            ['Location ID', 'Other systems', 'Additional notes', 'Last updated date']
            >>> lid.KEY
            'LocationID'
            >>> loc_codes_dat = loc_codes[lid.KEY]
            >>> type(loc_codes_dat)
            pandas.core.frame.DataFrame
            >>> loc_codes_dat.head()
                                          Location CRS  ... STANME_Note STANOX_Note
            0                 1999 Reorganisations      ...
            1                                   A1      ...
            2                       A463 Traded In      ...
            3  A483 Road Scheme Supervisors Closed      ...
            4                               Aachen      ...
            [5 rows x 12 columns]
        """

        verbose_ = collect_in_fetch_verbose(data_dir=dump_dir, verbose=verbose)

        loc_id_data = self.fetch_loc_id(update=update, verbose=verbose_)

        # Get other systems codes
        other_systems_codes = self.fetch_other_systems_codes(update=update, verbose=verbose_)

        # Get explanatory notes
        explanatory_notes = self.fetch_notes(update=update, verbose=verbose_)

        keys = [self.KEY, self.KEY_TO_OTHER_SYSTEMS, self.KEY_TO_NOTES]
        data_list = [loc_id_data, other_systems_codes, explanatory_notes]

        location_codes = {}
        for dat, key in zip(data_list, keys):
            if dat is None:
                location_codes.update({key: None})
            else:
                for k, v in dat.items():
                    if k != self.KEY_TO_LAST_UPDATED_DATE:
                        location_codes.update({k: v})

        # Get the latest updated date
        latest_update_date = max(d[self.KEY_TO_LAST_UPDATED_DATE] for d in filter(None, data_list))

        # Create a dict to include all information
        location_codes.update({self.KEY_TO_LAST_UPDATED_DATE: latest_update_date})

        if dump_dir:
            kwargs.update({'data': location_codes, 'data_name': self.KEY})
            self._save_data_to_file(dump_dir=dump_dir, verbose=verbose, **kwargs)

        return location_codes

    @staticmethod
    def _make_xref_dict(location_codes, keys, drop_duplicates=False, as_dict=False, main_key=None):
        # Extract relevant columns and remove empty rows
        key_locid = location_codes[['Location'] + keys].query(
            ' | '.join([f"{k} != ''" for k in keys]))

        # Further clean location_code
        if drop_duplicates:
            location_codes_ref = key_locid.drop_duplicates(keys, keep='first').set_index(keys)

        else:  # drop_duplicates is False or None
            locid_unique = key_locid.drop_duplicates(subset=keys, keep=False)

            dupl_temp_1 = key_locid[key_locid.duplicated(['Location'] + keys, keep=False)]
            dupl_temp_2 = key_locid[key_locid.duplicated(keys, keep=False)]
            duplicated_1 = dupl_temp_2[dupl_temp_1.eq(dupl_temp_2)].dropna().drop_duplicates()
            duplicated_2 = dupl_temp_2[~dupl_temp_1.eq(dupl_temp_2)].dropna()
            duplicated_entries = pd.concat([duplicated_1, duplicated_2], axis=0)

            grouped_duplicates = duplicated_entries.groupby(keys).agg(tuple)
            grouped_duplicates['Location'] = grouped_duplicates['Location'].map(
                lambda x: x[0] if len(set(x)) == 1 else x)

            locid_unique.set_index(keys, inplace=True)
            location_codes_ref = pd.concat([locid_unique, grouped_duplicates])

        if as_dict:
            location_codes_dict = location_codes_ref.to_dict()
            if main_key:
                location_codes_dict = {main_key: location_codes_dict.pop('Location')}
            else:
                location_codes_dict = location_codes_dict['Location']

        else:
            location_codes_dict = location_codes_ref

        return location_codes_dict

    def make_xref_dict(self, keys, initials=None, drop_duplicates=False, as_dict=False,
                       main_key=None, update=False, dump_it=False, dump_dir=None, verbose=False):
        """
        Creates a dictionary or dataframe containing location code data for the specified ``keys``.

        :param keys: A string or list of keys from ``['CRS', 'NLC', 'TIPLOC', 'STANOX', 'STANME']``
            representing the location codes.
        :type keys: str | list
        :param initials: A string or list of initials for which the codes are filtered;
            defaults to ``None``.
        :type initials: str | list | None
        :param drop_duplicates: If ``True``, removes duplicate entries from the result;
            defaults to ``False``.
        :type drop_duplicates: bool
        :param as_dict: If ``True``, returns the result as a dictionary;
            otherwise, returns a dataframe; defaults to ``False``.
        :type as_dict: bool
        :param main_key: The key used for the returned dictionary when ``as_dict=True``;
            defaults to ``None``.
        :type main_key: str | None
        :param update: Whether to check for updates to the package data; defaults to ``False``.
        :type update: bool
        :param dump_it: If ``True``, saves the result to a file; defaults to ``False``.
        :type dump_it: bool
        :param dump_dir: The directory path where the file can be saved, if ``dump_it=True``;
            defaults to ``None``.
        :type dump_dir: str | None
        :param verbose: Whether to print relevant information to the console; defaults to ``False``.
        :type verbose: bool | int
        :return: A dictionary or dataframe containing cross-reference location data
            for the specified ``keys``, or ``None`` if no data is available.
        :rtype: dict | pandas.DataFrame | None

        **Examples**::

            >>> from pyrcs.line_data import LocationIdentifiers
            >>> # from pyrcs import LocationIdentifiers
            >>> lid = LocationIdentifiers()
            >>> stanox_dict = lid.make_xref_dict(keys='STANOX')
            >>> type(stanox_dict)
            pandas.core.frame.DataFrame
            >>> stanox_dict.head()
                                      Location
            STANOX
            00005                       Aachen
            04309           Abbeyhill Junction
            04311        Abbeyhill Signal E811
            04308   Abbeyhill Turnback Sidings
            88601                   Abbey Wood
            >>> s_t_dictionary = lid.make_xref_dict(keys=['STANOX', 'TIPLOC'], initials='a')
            >>> type(s_t_dictionary)
            pandas.core.frame.DataFrame
            >>> s_t_dictionary.head()
                                              Location
            STANOX TIPLOC
            00005  AACHEN                       Aachen
            04309  ABHLJN           Abbeyhill Junction
            04311  ABHL811       Abbeyhill Signal E811
            04308  ABHLTB   Abbeyhill Turnback Sidings
            88601  ABWD                     Abbey Wood
            >>> ks = ['STANOX', 'TIPLOC']
            >>> ini = 'b'
            >>> main_k = 'Data'
            >>> s_t_dictionary = lid.make_xref_dict(ks, ini, main_k, as_dict=True)
            >>> type(s_t_dictionary)
            dict
            >>> list(s_t_dictionary.keys())
            ['Data']
            >>> list(s_t_dictionary['Data'].keys())[:5]
            [('55115', ''),
             ('23490', 'BABWTHL'),
             ('38306', 'BACHE'),
             ('66021', 'BADESCL'),
             ('81003', 'BADMTN')]
        """

        valid_keys = {'CRS', 'NLC', 'TIPLOC', 'STANOX', 'STANME'}
        if isinstance(keys, str):
            keys = [keys]
        if not all(k in valid_keys for k in keys):
            raise ValueError(f"`keys` must be one of {valid_keys}, but got {keys}.")

        if main_key and not isinstance(main_key, str):
            raise TypeError("`main_key` must be a string.")

        # Validate `initials`
        if initials:
            if isinstance(initials, str):
                initials = [validate_initial(initials, as_is=True)]
            elif not all(isinstance(x, str) and x in string.ascii_letters for x in initials):
                raise ValueError("`initials` must be a string or a list of letters.")

            dat_list = [
                self.fetch_loc_id(x, update=update, verbose=verbose).get(x.upper())
                for x in initials]
            location_codes = pd.concat(dat_list, ignore_index=True)

        else:
            location_codes = self.fetch_codes(update=update, verbose=verbose).get(self.KEY)

        if verbose == 2:
            print("Generating location code dictionary", end=" ... ")

        try:
            location_codes_dict = self._make_xref_dict(
                location_codes, keys=keys, drop_duplicates=drop_duplicates, as_dict=as_dict,
                main_key=main_key)

            if verbose == 2:
                print("Successfully.")

            if dump_it:
                self._save_data_to_file(
                    data=location_codes_dict,
                    data_name="-".join(keys) + (f"-{''.join(initials)}" if initials else ""),
                    ext=".json" if as_dict and len(keys) == 1 else ".pkl",
                    dump_dir=validate_dir(dump_dir) if dump_dir else self._cdd("xref-dicts"),
                    verbose=verbose)

            return location_codes_dict

        except Exception as e:
            _print_failure_message(e)
